# -*- coding: UTF-8 -*-

import os
from typing import Optional, Union, Tuple, Any

import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from pandas import DataFrame

from .. import util as ul
from ..util import path, collection, plot_end, plot_start

__name__: str = "plot_radar"


def radar(
    ax_x: collection,
    ax_y: collection,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    colors: collection = None,
    width: float = 4,
    height: float = 4,
    bottom: float = 0,
    center_text: str = None,
    rotation: float = 25,
    value_top: float = 0.1,
    text_top: float = 1.2,
    is_fixed: bool = False,
    is_angle: bool = True,
    y_limit: Tuple = (-0.5, 1),
    y_axis_scale: Tuple = (0, 1),
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):

    fig, ax = plot_start(width, height, bottom, output, show)

    ax_x = list(ax_x)
    ax_y = list(ax_y)

    # Create a circular bar chart
    theta = np.linspace(0, 2 * np.pi, len(ax_x), endpoint=False).tolist()
    ax_y += ax_y[:1]
    theta += theta[:1]

    width = 2 * 2.7 / len(ax_x)

    bars = ax.bar(theta, ax_y, width=width, color=colors, edgecolor='none', alpha=0.8, zorder=3, **kwargs)

    # Add category labels
    ax.set_xticks(theta)
    ax.set_xticklabels([])

    # Set y-axis range
    ax.set_ylim(y_limit[0], y_limit[1])

    # Remove the scale value of the circle
    ax.set_yticks(np.linspace(y_axis_scale[0], y_axis_scale[1], 6))  # Set the y-axis scale position
    ax.set_yticklabels([])  # Do not display scale values
    ax.set_theta_zero_location('N')  # Set polar axis position
    ax.set_theta_direction(-1)  # The angle increases counterclockwise

    # Add numerical labels
    for i, bar in enumerate(bars):
        height = bar.get_height()
        angle = np.degrees(theta[i])
        label_position = theta[i]
        ax.text(label_position, height + value_top if not is_fixed else value_top, round(height, 3), ha='center', va='center', color='#1f1f1f', rotation=-angle + rotation if is_angle else rotation)

    # Add radar line
    ax.plot(theta, ax_y, color='gray', linewidth=1, zorder=1)
    # Draw radar map
    ax.fill(theta, ax_y, color='#DDDDDD', alpha=0.1, zorder=2)

    if center_text is not None:
        ax.text(0, y_limit[0], center_text, ha='center', va='center', fontsize=14, color='black', zorder=11)

    # Set the y-axis scale line color to light gray
    ax.tick_params('y', colors='#DDDDDD', grid_alpha=0.6, zorder=8)

    # Set the color of the outermost circle line
    ax.spines['polar'].set_color('#DDDDDD')

    plt.grid(axis='x', linestyle='-', alpha=0.4, zorder=9)

    # Draw peripheral category labels
    for i, label in enumerate(ax_x):
        angle = np.degrees(theta[i])
        ax.text(theta[i], text_top, label, ha='center', va='center', color='#1f1f1f', zorder=20, rotation=-angle + rotation if is_angle else rotation)

    # Adjust the layout to prevent label overlap
    plt.tight_layout()

    plot_end(fig, title, x_name, y_name, output, show, close)


def radar_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    value: str = "rate",
    clusters: str = "clusters",
    color: Union[collection, str] = None,
    clusters_sort: Optional[list] = None,
    width: float = 4,
    height: float = 4,
    rotation: float = 65,
    title: str = None,
    value_top: float = 0.1,
    text_top: float = 1.2,
    is_fixed: bool = False,
    is_angle: bool = True,
    y_limit: Tuple = (-0.5, 1),
    y_axis_scale: Tuple = (0, 1),
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):

    def trait_plot(trait_: str, cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting bar {}".format(trait_))
        trait_score = cell_df_[cell_df_[trait_column_name] == trait_]

        # Sort
        if clusters_sort is not None:
            trait_score[clusters] = pd.Categorical(trait_score[clusters], categories=clusters_sort, ordered=True)
            trait_score = trait_score.sort_values(by=clusters)
            ax_x = clusters_sort
        else:
            trait_score = trait_score.sort_values([value], ascending=False)
            ax_x = trait_score[clusters].tolist()

        colors = None
        if color is not None:
            if isinstance(color, str):
                if color in trait_score.columns:
                    colors = trait_score[color]
            elif isinstance(color, collection):
                colors = color

        radar(
            ax_x=ax_x,
            ax_y=trait_score[value].tolist(),
            title=f"{title} {trait_}" if title is not None else title,
            colors=colors,
            width=width,
            height=height,
            rotation=rotation,
            value_top=value_top,
            text_top=text_top,
            is_fixed=is_fixed,
            is_angle=is_angle,
            y_limit=y_limit,
            y_axis_scale=y_axis_scale,
            center_text=trait_,
            output=os.path.join(output, f"{trait_}_enrichment_radar.pdf") if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    trait_list = list(set(trait_df[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )
        raise ValueError(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait_=trait, cell_df_=trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Union, Optional, Any, Literal

import numpy as np
import pandas as pd
from pandas import DataFrame

import seaborn as sns
from matplotlib import pyplot as plt
from statannotations.Annotator import Annotator

from .. import util as ul
from ..util import path, collection, plot_color_types, plot_start, plot_end

__name__: str = "plot_bar"


def bar(
    ax_x: collection,
    ax_y: collection,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    color: str = "#70b5de",
    text_color: str = "#000205",
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    text_left_move: float = 0.1,
    direction: Literal['vertical', 'horizontal'] = "vertical",
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    fig, ax = plot_start(width, height, bottom, output, show)

    ax_x = np.array(ax_x).astype(str)

    if direction == 'vertical':
        ax.bar(ax_x, ax_y, color=color, **kwargs)
    elif direction == 'horizontal':
        ax.barh(ax_x, ax_y, color=color, **kwargs)
    else:
        ul.log(__name__).error("The `direction` must be 'vertical' or 'horizontal'.")
        raise ValueError("The `direction` must be 'vertical' or 'horizontal'.")

    ax.set_xticklabels(labels=list(ax_x), rotation=65)

    # Draw numerical values
    for i, v in enumerate(list(ax_y)):
        plt.text(
            x=i - text_left_move,
            y=0.03 if v < 0.03 else v / 2,
            s=str(round(v, 3)),
            rotation=90,
            color=text_color
        )

    plot_end(fig, title, x_name, y_name, output, show, close)


def two_bar(
    ax_x: collection,
    ax_y: Tuple,
    x_name: str = None,
    y_name: str = None,
    legend: Tuple = ("1", "2"),
    color: Tuple = ("#2e6fb7", "#f7f7f7"),
    text_color: str = "#000205",
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    rotation: float = 65,
    text_left_move: float = 0.15,
    y_limit: Tuple[float, float] = (0, 1),
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):

    fig, ax = plot_start(width, height, bottom, output, show)

    ax_x = np.array(ax_x).astype(str)
    ax.bar(ax_x, ax_y[0], label=legend[0], color=color[0], **kwargs)
    ax.bar(ax_x, ax_y[1], bottom=ax_y[0], label=legend[1], color=color[1], **kwargs)

    ax.legend()

    ax.set_ylim(y_limit)

    ax.set_xticks(range(len(ax_x)))
    ax.set_xticklabels(labels=list(ax_x), rotation=rotation)

    # Draw numerical values
    for i, v in enumerate(list(ax_y[0])):
        plt.text(
            x=i - text_left_move,
            y=0.03 if v < 0.03 else v / 2,
            s=str(round(v, 3)),
            rotation=90,
            color=text_color
        )

    for spine in ["top", "left", "right", "bottom"]:
        ax.spines[spine].set_linewidth(1)

    ax.spines['bottom'].set_linewidth(1)
    ax.grid(axis='y', ls='--', c='gray')
    ax.set_axisbelow(True)

    plot_end(fig, title, x_name, y_name, output, show, close)


def class_bar(
    df: DataFrame,
    value: str = "rate",
    by: str = "value",
    clusters: str = "clusters",
    color: Tuple = ("#2e6fb7", "#f7f7f7"),
    x_name: str = "Cell type",
    y_name: str = "Enrichment ratio",
    legend: Tuple = ("Enrichment", "Conservative"),
    text_color: str = "#000205",
    clusters_sort: Optional[list] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    rotation: float = 65,
    title: str = None,
    text_left_move: float = 0.15,
    y_limit: Tuple[float, float] = (0, 1),
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):

    df1 = df[df[by] == 1]
    df2 = df[df[by] == 0]

    # Sort
    if clusters_sort is not None:
        df1[clusters] = pd.Categorical(df1[clusters], categories=clusters_sort, ordered=True)
        df1 = df1.sort_values(by=clusters)
        df2[clusters] = pd.Categorical(df2[clusters], categories=clusters_sort, ordered=True)
        df2 = df2.sort_values(by=clusters)
        ax_x = clusters_sort
    else:
        df1 = df1.sort_values([value], ascending=False)
        df2 = df2.sort_values([value])
        ax_x = df1[clusters]

    ax_y = (df1[value], df2[value])

    two_bar(
        ax_x=ax_x,
        ax_y=ax_y,
        x_name=x_name,
        y_name=y_name,
        legend=legend,
        width=width,
        height=height,
        color=color,
        text_color=text_color,
        bottom=bottom,
        rotation=rotation,
        text_left_move=text_left_move,
        y_limit=y_limit,
        title=title,
        output=output,
        show=show,
        close=close,
        **kwargs
    )


def bar_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    value: str = "rate",
    clusters: str = "clusters",
    x_name: str = "Cell type",
    y_name: str = "Enrichment ratio",
    color: Tuple = ("#2e6fb7", "#f7f7f7"),
    legend: Tuple = ("Enrichment", "Conservative"),
    text_color: str = "#000205",
    clusters_sort: Optional[list] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    rotation: float = 65,
    title: str = None,
    text_left_move: float = 0.15,
    y_limit: Tuple[float, float] = (0, 1),
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):
    def trait_plot(trait_: str, cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting bar {}".format(trait_))
        # get gene score
        trait_score = cell_df_[cell_df_[trait_column_name] == trait_]
        # Sort gene scores from small to large
        class_bar(
            df=trait_score,
            value=value,
            clusters=clusters,
            title=f"{title} {trait_}" if title is not None else title,
            color=color,
            legend=legend,
            width=width,
            x_name=x_name,
            y_name=y_name,
            clusters_sort=clusters_sort,
            height=height,
            bottom=bottom,
            rotation=rotation,
            text_left_move=text_left_move,
            y_limit=y_limit,
            text_color=text_color,
            output=os.path.join(output, f"cell_{trait_}_enrichment_bar.pdf") if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    trait_list = list(set(trait_df[trait_column_name]))

    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )
        raise ValueError(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )

    # plot
    if trait_name == "All":

        for trait in trait_list:
            trait_plot(trait_=trait, cell_df_=trait_df)

    else:
        trait_plot(trait_name, trait_df)


def bar_significance(
    df: DataFrame,
    x: str,
    y: str,
    hue: str,
    x_name: str = None,
    y_name: str = None,
    anchor: str = None,
    legend: str = None,
    legend_list: list = None,
    hue_order: list = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    legend_gap: float = 1.15,
    line_width: float = 0.5,
    capsize: float = 0.1,
    errcolor: str = "k",
    start_color_index: int = 0,
    color_step_size: int = 0,
    color_type: str = "set",
    test: str = "t-test_ind",
    ci: Union[str, float] = "sd",
    x_rotation: float = 0,
    x_deviation: float = 0.02,
    y_deviation: float = 0.02,
    y_limit: Tuple[float, float] = (0, 1),
    anno: bool = False,
    anno_fontsize: float = 7,
    line_height: float = 0.01,
    line_offset: float = 0.01,
    colors: Union[list, dict] = None,
    title: str = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    """
    A bar chart with specified anchor significance.
    :param df: Input DataFrame containing the data to plot
    :param x: Column name for x-axis categories
    :param y: Column name for y-axis values
    :param hue: Column name for grouping bars by color
    :param x_name: Label for x-axis (optional)
    :param y_name: Label for y-axis (optional)
    :param anchor: Reference group for pairwise significance testing
    :param hue_order: Order of hue categories
    :param legend: Legend title (default: "category")
    :param legend_list: Subset of hue values to include in plot
    :param width: Figure width in inches
    :param height: Figure height in inches
    :param bottom: Figure bottom in inches
    :param legend_gap: Vertical gap between plot and legend
    :param line_width: Width of error bars and significance lines
    :param start_color_index: Starting index in color palette
    :param color_step_size: Step size when cycling through palette
    :param color_type: Name of seaborn palette to use
    :param test: Statistical test for pairwise comparisons
        {"t-test_ind", "t-test_welch", "t-test_paired", "Mann-Whitney", "Mann-Whitney-gt", "Mann-Whitney-ls",
         "Levene", "Wilcoxon", "Kruskal", "Brunner-Munzel"}
    :param ci: Confidence interval type or value
    :param capsize: Width of the error-bar caps
    :param errcolor: Color of the error bars
    :param line_offset: Vertical offset for significance lines
    :param line_height: Height of significance lines
    :param x_rotation: Rotation angle for x-axis tick labels
    :param x_deviation: Horizontal offset for bar annotations
    :param y_deviation: Vertical offset for bar annotations
    :param y_limit: Tuple setting y-axis limits
    :param anno: Whether to annotate bars with their values
    :param anno_fontsize: Font size for bar annotations
    :param colors: Custom color list (overrides palette)
    :param title: Plot title
    :param output: Path to save figure (optional)
    :param show: Whether to display the plot
    :param close:
    :return: None
    """

    fig, ax = plot_start(width, height, bottom, output, show)

    if legend_list is not None:
        new_data: DataFrame = df[df[hue].isin(legend_list)].copy()
    else:
        new_data: DataFrame = df.copy()

    if legend is None:
        legend = "category"

    new_data.loc[:, legend] = new_data[hue].astype(str)

    new_data_columns = list(new_data.columns)

    if hue_order is not None:
        # Sort
        new_data[legend] = pd.Categorical(new_data[legend], categories=hue_order, ordered=True)
        new_data = new_data.sort_values(by=legend)

    hue_types = new_data[legend].unique().tolist()

    if colors is not None:
        if isinstance(colors, list):
            palette = colors
        elif isinstance(colors, dict):
            palette = []
            for hue_type in hue_types:
                if hue_type in colors:
                    palette.append(colors[hue_type])
                else:
                    ul.log(__name__).warning(f"`{hue_type}` is not in `colors` ({colors})")
                    raise ValueError(f"`{hue_type}` is not in `colors` ({colors})")
        else:
            ul.log(__name__).error(f"`colors` ({colors}) must be a list or dict")
            raise ValueError(f"`colors` ({colors}) must be a list or dict")
    else:
        if "color" in new_data_columns:
            palette = new_data["color"]
        else:
            palette = []

            for i in range(len(hue_types)):
                palette.append(plot_color_types[color_type][start_color_index + i * color_step_size + i])

    # Set y-axis limits first to prevent seaborn from overriding
    ax.set_ylim(y_limit)
    # Draw barplot, note ax receives return value to keep handles
    ax = sns.barplot(
        data=new_data,
        x=x,
        y=y,
        hue=legend,
        hue_order=hue_order,
        errorbar=('ci', ci),
        capsize=capsize,
        err_kws={'color': errcolor, 'linewidth': line_width},
        ax=ax,
        palette=palette,
        edgecolor=errcolor,
        linewidth=line_width,
        **kwargs
    )

    if anno:
        for p in ax.patches:
            y_value = p.get_height()
            height = p.get_height() / 2 - y_deviation
            height = 0.03 if height < 0.03 else height
            x = p.get_x() + p.get_width() / 2 + x_deviation
            ax.annotate(
                f'{y_value:.2f}',
                (x, height),
                textcoords="offset points",
                ha='center',
                va='bottom',
                rotation=90,
                fontsize=anno_fontsize
            )

    if anchor is not None:

        # Add p value
        box_pairs: list = []

        x_list = new_data[x].unique().tolist()
        class_list = new_data[legend].unique().tolist()

        if anchor not in class_list:
            ul.log(__name__).error(f"`anchor` ({anchor}) is not in the `df[hue]` ({class_list})")
            raise ValueError(f"`anchor` ({anchor}) is not in the `df[hue]` ({class_list})")

        class_list.remove(anchor)

        for x_ele in x_list:

            for class_ele in class_list:
                box_pairs.append(((x_ele, anchor), (x_ele, class_ele)))

        annotator = Annotator(ax=ax, data=new_data, x=x, y=y, hue=legend, hue_order=hue_order, pairs=box_pairs)
        annotator.configure(test=test, text_format='star', line_height=line_height, line_offset=line_offset,
                            line_width=0.7)
        annotator.apply_and_annotate()

    ax.tick_params(which='major', direction='in', length=3, width=1.0, bottom=False)

    for spine in ["top", "left", "right"]:
        ax.spines[spine].set_visible(False)

    ax.spines['bottom'].set_linewidth(1)
    ax.grid(axis='y', ls='--', c='gray')
    ax.set_axisbelow(True)

    if x_rotation != 0:
        ax.tick_params(axis='x', rotation=x_rotation)

    plt.legend(loc='upper left', bbox_to_anchor=(0.0, legend_gap), ncol=2)

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

import math
import os
import random
import string
import threading
import time
from functools import wraps
from typing import Tuple, Union, Literal, Callable, Any
import psutil

import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from numpy import asarray
from anndata import AnnData
from pandas import DataFrame
from scipy import sparse

from yzm_file import StaticMethod
from yzm_log import Logger

from .. import util as ul
from ._constant_ import dense_data, sparse_data, sparse_matrix, matrix_data, number, collection, project_name

__name__: str = "util_core"


def file_method(name: str = None) -> StaticMethod:
    name = f"{project_name}_{name}" if name is not None else project_name
    return StaticMethod(log_file=os.path.join(ul.log_file_path, name), is_form_log_file=ul.is_form_log_file)


def log(name: str = None) -> Logger:
    name = f"{project_name}_{name}" if name is not None else project_name
    return Logger(name, log_path=os.path.join(ul.log_file_path, name), is_form_file=ul.is_form_log_file)


def track_with_memory(interval: float = 60) -> Callable:
    """
    Decorator: Records memory usage at fixed intervals during function execution and returns the result, elapsed time, and memory list.

    Parameters
    ----------
    interval : float, optional
        Sampling interval (seconds), default is 60 seconds.

    Returns
    -------
    Callable
        Decorator function; when the wrapped function is called, it returns a dictionary containing:
        - 'result': the original function's return value
        - 'time': function execution time (seconds) if is_monitor is True, otherwise None.
        - 'memory': list of sampled memory usage (bytes) if is_monitor is True, otherwise None.
    """

    def decorator(func) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Union[Any, dict]:

            process = psutil.Process(os.getpid())

            stop_monitor = False

            mem_list = []

            def monitor():
                nonlocal stop_monitor

                while not stop_monitor:
                    current_mem = process.memory_info().rss
                    mem_list.append(current_mem)

                    time.sleep(interval)

            t = threading.Thread(target=monitor, daemon=True)
            t.start()

            start_time = time.perf_counter()
            _result_ = func(*args, **kwargs)
            end_time = time.perf_counter()

            stop_monitor = True
            t.join()

            exec_time = end_time - start_time

            return {
                'result': _result_,
                'time': exec_time,
                'memory': mem_list
            }

        return wrapper

    return decorator


def to_dense(sm: matrix_data, is_array: bool = False) -> dense_data:
    """
    Convert sparse matrix to dense matrix
    :param sm: sparse matrix
    :param is_array: True converted to array form, False natural output
    :return: dense matrix
    """

    if sm is None:
        log(__name__).warning("The input matrix (sm parameter) is not feasible")
        return np.asmatrix([])

    if isinstance(sm, sparse_data):
        return np.array(sm.todense()) if is_array else sm.todense()

    dense_sm = sm.todense() if sparse.issparse(sm) else sm
    return np.array(dense_sm) if is_array else dense_sm


def to_sparse(dm: dense_data, way_callback=sparse.csr_matrix, is_matrix: bool = True) -> sparse_matrix:
    """
    Convert dense matrix to sparse matrix
    :param dm: dense matrix
    :param way_callback: How to form sparse matrix
    :param is_matrix: True converted to matrix form, False natural output
    :return: sparse matrix
    """

    if dm is None:
        log(__name__).warning("The input matrix (dm parameter) is not feasible")
        return sparse.coo_matrix([])

    if isinstance(dm, dense_data):
        return way_callback(dm) if is_matrix else dm

    sparse_m = dm if sparse.issparse(dm) else way_callback(dm)
    return way_callback(sparse_m) if is_matrix else sparse_m


def sum_min_max(data: matrix_data, axis: int = 1) -> Tuple[number, number]:
    """
    Obtain the minimum/maximum sum of rows in the matrix
    :param data: matrix data
    :param axis: {0, 1} 1: row, 0: col
    :return: Minimum value of rows, maximum value of rows
    """
    rows, cols = data.shape

    if rows == 0 or cols == 0:
        return 0, 0

    rows_sum = list(np.array(data.sum(axis=axis)).flatten())
    return min(rows_sum), max(rows_sum)


def get_index(position: number, positions_list: list, is_sort: bool = True) -> Union[int, Tuple[int, int]]:
    """
    Search for position information. Similar to half search.
        If the position exists in the list, return the index.
        If it does not exist, return the index located between the two indexes
    :param position: position
    :param positions_list: position list
    :param is_sort: True
    :return: position index
    """

    if is_sort:
        positions_list.sort()

    # search
    position_size: int = len(positions_list)
    left, right = 0, position_size - 1

    while left <= right:
        mid = (left + right) // 2

        if positions_list[mid] == position:
            return mid
        elif positions_list[mid] > position:
            right = mid - 1
        else:
            left = mid + 1

    return right, left


def list_duplicate_set(data: list) -> list:
    """
    Append numbering to duplicate information
    :param data: input data
    :return: Unique data with constant quantity
    """

    if len(data) == len(set(data)):
        return data

    new_data = []
    range_data = range(len(data))

    is_warn: bool = False

    for i in range_data:

        # judge duplicate
        if data[i] not in new_data:
            new_data.append(data[i])
        else:
            j: int = 2

            while True:

                if not isinstance(data[i], str) and not is_warn:
                    log(__name__).warning("Convert non string types to string types.")
                    is_warn = True

                # format new index
                elem: str = data[i] + "+" + str(j)

                # Add index
                if elem not in new_data:
                    new_data.append(elem)
                    break

                j += 1

    return new_data


def get_sub_data(data: collection, size: int) -> collection:
    # get information
    old_size = len(data)

    if size >= old_size:
        log(__name__).warning("The given size is greater than the original data size")
        return data

    old_data: list = data.copy()
    rate = size / old_size
    # add sub data
    new_data: list = []

    for i in range(size):
        new_data.append(old_data[math.floor(rate * i)])

    return new_data


def split_matrix(data: matrix_data, axis: Literal[0, 1] = 0, chunk_number: int = 1000) -> list:
    # get size
    new_data = to_dense(data, is_array=True)
    rows, cols = new_data.shape

    # get number
    total: int = rows if axis == 0 else cols

    # get split number
    split_number = total // chunk_number

    # Determine whether to divide equally
    tail_number = total % chunk_number
    if tail_number != 0:
        split_number += 1

    log(__name__).info(f"Divide the matrix into {split_number} parts")

    # Add data
    split_data_list = []
    for i in range(split_number):
        # set index
        start_index: int = i * chunk_number
        end_index: int = total if (i + 1) * chunk_number > total else (i + 1) * chunk_number
        split_data = new_data[start_index:end_index, :] if axis == 0 else new_data[:, start_index:end_index]
        split_data_list.append(split_data)

    return split_data_list


def merge_matrix(datas: list, axis: Literal[0, 1] = 0) -> list:
    # get size
    size = len(datas)
    range_size = range(size)

    # get row col
    constant: int = datas[0].shape[1] if axis == 0 else datas[0].shape[0]
    total: int = 0
    shapes: list = []

    # get chunk size
    for i in range_size:
        shape = datas[i].shape
        judge = shape[1] if axis == 0 else shape[0]

        # judge size
        if judge != constant:
            log(__name__).error("Inconsistent traits in the input dataset set.")
            raise ValueError("Inconsistent traits in the input dataset set.")

        total += shape[0] if axis == 0 else shape[1]
        shapes.append(shape)

    # format matrix
    matrix_shape = (total if axis == 0 else constant, constant if axis == 0 else total)
    log(__name__).info(f"Merge matrix {matrix_shape} shape")
    matrix: matrix_data = np.zeros(matrix_shape)

    # merge matrix
    col_record: int = 0
    for i in range_size:
        col_i = datas[i].shape[0] if axis == 0 else datas[i].shape[1]
        if axis == 0:
            matrix[col_record:col_record + col_i, :] = datas[i]
        else:
            matrix[:, col_record:col_record + col_i] = datas[i]
        col_record += col_i

    return matrix


def list_index(data: list) -> Tuple[list, collection]:
    info: list = []

    size: int = len(data)

    types: asarray = np.unique(data)

    for type_ in types:
        type_info: list = []
        for i in range(size):
            if type_ == data[i]:
                type_info.append(i)
        info.append(set(type_info))

    return info, types


def numerical_bisection_step(min_value: float, max_value: float, step_length: float) -> Tuple[collection, int]:
    if min_value > max_value:
        log(__name__).error(f"`min_value` ({min_value}) must be smaller than `max_value` ({max_value}).")
        raise ValueError(f"`min_value` ({min_value}) must be smaller than `max_value` ({max_value}).")

    number_list: list = []

    i = 0
    while min_value <= max_value:
        number_list.append(min_value)
        i += 1
        min_value += step_length

    return number_list, i


def get_real_predict_label(
    df: DataFrame,
    map_cluster: Union[str, collection],
    clusters: str = "clusters",
    value: str = "value"
) -> Tuple[DataFrame, int, list]:
    df_sort: DataFrame = df.sort_values([value], ascending=False)

    # Obtain the type of positive set clustering corresponding to the trait
    cluster_list: list = []
    if isinstance(map_cluster, str):
        cluster_list.append(map_cluster)
    else:
        cluster_list = list(map_cluster)

    # total label size
    total_size = df.shape[0]

    # true label size
    df_sort.insert(0, "true_label", 0)
    df_sort.loc[df_sort[df_sort[clusters].isin(cluster_list)].index, "true_label"] = 1

    # predict label size
    df_cluster = df[df[clusters].isin(cluster_list)].copy()
    df_cluster_size = df_cluster.shape[0]
    predict_label = list(np.ones(df_cluster_size))
    predict_label.extend(np.zeros(total_size - df_cluster_size))
    df_sort.insert(0, "predict_label", 0)
    df_sort.loc[:, "predict_label"] = predict_label
    df_sort["predict_label"] = df_sort["predict_label"].astype(int)

    return df_sort, df_cluster_size, cluster_list


def strings_map_numbers(str_list: list, start: int = 0) -> list:
    # Create an empty dictionary to store the mapping of strings to numerical values
    mapping = {}

    # Traverse the list and assign a unique numerical value to each string
    for i, item in enumerate(set(str_list), start=start):  # 使用set去除重复项，并从1开始编号
        mapping[item] = i

    # Use list derivation to convert strings to corresponding numerical values
    numeric_list = [mapping[item] for item in str_list]

    return numeric_list


def generate_str(length: int = 10) -> str:
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


def set_inf_value(matrix: matrix_data) -> None:
    # solve -Inf/Inf value
    matrix_inf = np.logical_and(np.isinf(matrix), matrix > 0)
    matrix__inf = np.logical_and(np.isinf(matrix), matrix < 0)

    # set inf
    if np.any(matrix_inf):
        matrix[matrix_inf] = np.max(matrix[~matrix_inf]) * 2

    # set -inf
    if np.any(matrix__inf):
        matrix[matrix__inf] = np.min(matrix[~matrix__inf]) / 2


def check_adata_get(adata: AnnData, layer: str = None, is_dense: bool = True, is_matrix: bool = False) -> AnnData:
    # judge input data
    if adata.shape[0] == 0:
        log(__name__).warning("Input data is empty")
        raise ValueError("Input data is empty")

    # get data
    data: AnnData = adata.copy()

    # judge layers
    if layer is not None:
        if layer not in list(data.layers):
            log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
            raise ValueError("The value of the `layer` parameter must be one of the keys in `adata.layers`.")

        data.X = to_dense(data.layers[layer], is_array=True) if is_dense \
            else to_sparse(data.layers[layer], is_matrix=is_matrix)
    else:
        data.X = to_dense(data.X, is_array=True) if is_dense else to_sparse(data.X, is_matrix=is_matrix)

    return data


def add_cluster_info(data: DataFrame, data_ref: DataFrame, cluster: str) -> DataFrame:
    new_data: DataFrame = data.copy()
    if data_ref is not None and cluster not in new_data.columns:

        new_data: DataFrame = pd.merge(new_data, data_ref, how="left", left_index=True, right_index=True)

        if "barcode_x" in new_data.columns:
            new_data["barcode"] = new_data["barcode_x"]
            new_data.drop("barcode_x", axis=1, inplace=True)

            if "barcode_y" in new_data.columns:
                new_data.drop("barcode_y", axis=1, inplace=True)

        if "barcodes_x" in new_data.columns:
            new_data["barcodes"] = new_data["barcodes_x"]
            new_data.drop("barcodes_x", axis=1, inplace=True)

            if "barcodes_y" in new_data.columns:
                new_data.drop("barcodes_y", axis=1, inplace=True)

    if cluster not in new_data.columns:
        log(__name__).error(f"`{cluster}` is not in `adata.obs.columns`.")
        raise ValueError(f"`{cluster}` is not in `columns` ({new_data.columns}).")

    return new_data


def check_gpu_availability(verbose: bool = True) -> bool:
    available = torch.cuda.is_available()

    if verbose:

        if available:
            log(__name__).info("GPU is available.")
            log(__name__).info(f"Number of GPUs: {torch.cuda.device_count()}")
            log(__name__).info(f"GPU Name: {torch.cuda.get_device_name(torch.cuda.current_device())}")
        else:
            log(__name__).info("GPU is not available.")

    return available


def plot_start(
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    output: str = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    fig, ax = plt.subplots(figsize=(width, height))
    fig.subplots_adjust(bottom=bottom)

    return fig, ax


def plot_end(
    fig,
    title: str = None,
    x_name: str = None,
    y_name: str = None,
    output: str = None,
    show: bool = True,
    close: bool = False,
    dpi: float = 300
):
    if title is not None:
        plt.title(title)

    if x_name is not None:
        plt.xlabel(x_name, rotation=0)

    if y_name is not None:
        plt.ylabel(y_name, rotation=90)

    if output is not None:

        if output.endswith(".pdf"):

            with PdfPages(output) as pdf:
                pdf.savefig(fig)

        elif output.endswith(".png") or output.endswith(".jpg") or output.endswith(".svg"):
            plt.savefig(output, dpi=dpi)
        else:
            plt.savefig(f"{output}.png", dpi=dpi)

    if show:
        plt.show()

    if close:
        plt.close('all')


def generate_hex_colors(num_colors):
    colors = []

    while len(colors) < num_colors:
        color = "#{:02X}{:02X}{:02X}".format(
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255)
        )
        colors.append(color)

    return colors



# -*- coding: UTF-8 -*-

import time
import warnings
from typing import Optional, Literal

import numpy as np
from anndata import AnnData
from pandas import DataFrame

from .. import util as ul
from ..file import save_h5ad
from ..tool import lsi, tf_idf
from ..util import add_cluster_info, matrix_data, set_inf_value, collection

__name__: str = "preprocessing_scanpy"

_Method = Optional[Literal['logreg', 't-test', 'wilcoxon', 't-test_overestim_var']]


def filter_data(
    adata: AnnData,
    min_cells: int = 1,
    min_peaks: int = 1,
    min_peaks_counts: int = 1,
    min_cells_counts: int = 1,
    cell_rate: Optional[float] = None,
    peak_rate: Optional[float] = None,
    is_copy: bool = False,
    is_min_cell: bool = True,
    is_min_peak: bool = False
) -> AnnData:
    """
    Filter scATAC-seq data
    :param adata: scATAC-seq data
    :param min_peaks_counts: Minimum number of counts required for a peak to pass filtering
    :param min_cells: Minimum number of cells expressed required for a peak to pass filtering
    :param min_cells_counts: Minimum number of counts required for a cell to pass filtering
    :param min_peaks: Minimum number of peaks expressed required for a cell to pass filtering
    :param cell_rate: Removing the percentage of cell count in total cell count only takes effect when the min_cells
        parameter is None
    :param peak_rate: Removing the percentage of peak count in total peak count only takes effect when the min_peaks
        parameter is None
    :param is_copy: Do you want to deeply copy data
    :param is_min_cell: Whether to screen cells
    :param is_min_peak: Whether to screen peaks
    :return: scATAC-seq data
    """

    # start time
    start_time = time.perf_counter()

    import scanpy as sc

    ul.log(__name__).info("Filter scATAC-seq data")

    if adata.shape[0] == 0:
        ul.log(__name__).error("The scATAC data is empty")
        raise ValueError("The scATAC data is empty")

    filter_adata = adata.copy() if is_copy else adata
    cells_count, peaks_count = filter_adata.shape

    if cell_rate is not None:

        if cell_rate <= 0 or cell_rate >= 1:
            ul.log(__name__).error("The parameter of `cell_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `cell_rate` should be between 0 and 1.")

        _min_cells_ = int(filter_adata.shape[0] * cell_rate)

        if _min_cells_ > 1:
            min_cells = _min_cells_

    if peak_rate is not None:

        if peak_rate <= 0 or peak_rate >= 1:
            ul.log(__name__).error("The parameter of `peak_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `peak_rate` should be between 0 and 1.")

        _min_peaks_ = int(filter_adata.shape[1] * peak_rate)

        if _min_peaks_ > 1:
            min_peaks = _min_peaks_

    ul.log(__name__).info(f"min cells: {min_cells}, min peaks: {min_peaks}")
    sc.pp.filter_genes(filter_adata, min_cells=min_cells)
    sc.pp.filter_cells(filter_adata, min_genes=min_peaks)

    # filter peaks and cell
    if is_min_peak:
        sc.pp.filter_genes(filter_adata, min_counts=min_peaks_counts)

    if is_min_cell:
        sc.pp.filter_cells(filter_adata, min_counts=min_cells_counts)
    else:
        sc.pp.filter_cells(filter_adata)

    # judge cells count
    if filter_adata.shape[0] == 0:
        ul.log(__name__).error(
            "After screening, the number of cells was 0. Suggest setting the `is_min_peak` parameter to `False` or "
            "lowering the `cell_rate` and `peak_rate` parameters to try again"
        )
        return filter_adata

    ul.log(__name__).info(
        f"Filtered out cells {cells_count - filter_adata.shape[0]}, "
        f"Filtered out peaks {peaks_count - filter_adata.shape[1]}"
    )
    ul.log(__name__).info(f"Size of filtered scATAC-seq data: {filter_adata.shape}")
    filter_adata.uns["step"] = 0
    filter_adata.uns["elapsed_time"] = time.perf_counter() - start_time

    return filter_adata


def get_difference_genes(
    adata: AnnData,
    cluster: str,
    method: _Method = "wilcoxon",
    cell_anno: Optional[DataFrame] = None,
    diff_genes_file: Optional[str] = None
) -> AnnData:

    import scanpy as sc

    # add cell annotation information
    adata.obs = add_cluster_info(adata.obs, cell_anno, cluster)

    if "log1p" not in adata.uns_keys():
        ul.log(__name__).info("The `log1p` not detected in `adata.uns_keys`, `log1p` operation needs to be performed.")
        raise ValueError("The `log1p` not detected in `adata.uns_keys`, `log1p` operation needs to be performed.")

    if "base" not in adata.uns["log1p"].keys():
        adata.uns["log1p"].update({"base": None})

    # gene
    ul.log(__name__).info("Rank genes for characterizing groups.")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        sc.tl.rank_genes_groups(adata=adata, groupby=cluster, method=method, use_raw=False)

    # get difference genes for each `cluster`
    diff_genes = adata.uns['rank_genes_groups']['names']

    # gene names
    gene_list: list = list(adata.var.index)
    gene_list.sort()
    gene_dict: dict = dict(zip(gene_list, range(len(gene_list))))

    # obs
    cluster_info: DataFrame = adata.obs.copy().groupby([cluster], as_index=False).size()
    cluster_info.index = cluster_info[cluster].astype(str)

    cluster_list: list = cluster_info.index.tolist()
    cluster_list.sort()

    _shape_ = (adata.shape[1], cluster_info.shape[0])
    diff_genes_score_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_p_value_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_adjusted_p_value_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_log2_fold_change_matrix: matrix_data = np.zeros(_shape_)
    del _shape_

    # cluster
    for _cluster_ in cluster_list:
        ul.log(__name__).info(f"Obtaining differentially expressed genes for `cluster` ({_cluster_}).")
        # obtain cluster difference gene data
        _cluster_data_: DataFrame = sc.get.rank_genes_groups_df(adata, group=_cluster_)
        _cluster_index_: int = cluster_list.index(_cluster_)

        # Add data value
        for _gene_name_, _score_, _p_value_, _adjusted_p_value_, _log2_fold_change_ in zip(
            _cluster_data_["names"],
            _cluster_data_["scores"],
            _cluster_data_["pvals"],
            _cluster_data_["pvals_adj"],
            _cluster_data_["logfoldchanges"]
        ):
            _gene_index_: int = gene_dict[_gene_name_]
            diff_genes_score_matrix[_gene_index_, _cluster_index_] = 0 if np.isnan(_score_) else _score_
            diff_genes_p_value_matrix[_gene_index_, _cluster_index_] = 1 if np.isnan(_p_value_) else _p_value_
            diff_genes_adjusted_p_value_matrix[_gene_index_, _cluster_index_] = 1 if np.isnan(_adjusted_p_value_) else _adjusted_p_value_
            diff_genes_log2_fold_change_matrix[_gene_index_, _cluster_index_] = 0 if np.isnan(_log2_fold_change_) else _log2_fold_change_

        del _cluster_data_, _cluster_index_

    set_inf_value(diff_genes_score_matrix)
    set_inf_value(diff_genes_p_value_matrix)
    set_inf_value(diff_genes_adjusted_p_value_matrix)
    set_inf_value(diff_genes_log2_fold_change_matrix)

    diff_genes_p_value_matrix[diff_genes_p_value_matrix == 0] = np.min(diff_genes_p_value_matrix[diff_genes_p_value_matrix != 0])
    diff_genes_adjusted_p_value_matrix[diff_genes_adjusted_p_value_matrix == 0] = np.min(diff_genes_adjusted_p_value_matrix[diff_genes_adjusted_p_value_matrix != 0])

    # create
    diff_genes_adata: AnnData = AnnData(diff_genes_score_matrix, obs=adata.var, var=cluster_info)
    diff_genes_adata.layers["p_value"] = diff_genes_p_value_matrix
    diff_genes_adata.layers["adjusted_p_value"] = diff_genes_adjusted_p_value_matrix
    diff_genes_adata.layers["log2_fold_change"] = diff_genes_log2_fold_change_matrix

    # Add diff_genes
    diff_genes_adata.uns["diff_genes"] = diff_genes

    if diff_genes_file is not None:
        save_h5ad(diff_genes_adata, diff_genes_file)

    return diff_genes_adata


def paga_trajectory(
    adata: AnnData,
    layer: Optional[str] = None,
    latent: str = "X_pca",
    groups: str = "louvain",
    position: Optional[collection] = None,
    lsi_components: int = 50,
    root_cluster: Optional[str] = None,
    n_neighbors: int = 15,
    resolution: float = 1.0,
    is_denoise: bool = True,
) -> None:
    import scanpy as sc

    if position is not None:

        if len(position) != 2:
            ul.log(__name__).error("The `position` parameter must contain two elements, for example: `(UMAP1, UMAP2)`.")
            raise ValueError("The `position` parameter must contain two elements, for example: `(UMAP1, UMAP2)`.")

        if position[0] not in adata.obs.columns or position[1] not in adata.obs.columns:
            ul.log(__name__).error("The value in the `position` parameter must be one of the column names in `adata.obs`.")
            raise ValueError("The value in the `position` parameter must be one of the column names in `adata.obs`.")

    fixed_name: str = "X_pca"

    is_run: bool = latent not in adata.obsm

    status: int = 0

    if is_run:

        if layer is None:
            counts = adata.X
        else:

            if layer not in adata.layers:
                ul.log(__name__).error("The value of the `layer` parameter must be one of the keys in `adata.layers`.")
                raise ValueError("The value of the `layer` parameter must be one of the keys in `adata.layers`.")

            counts = adata.layers[layer]

        tf_idf_matrix = tf_idf(counts)
        del counts

        lsi_matrix = lsi(tf_idf_matrix, n_components=lsi_components)
        adata.obsm['X_pca'] = lsi_matrix
        del tf_idf_matrix, lsi_matrix

        status = 1

    else:

        if latent != fixed_name:
            adata.obsm['X_pca'] = adata.obsm[latent]
            status = 2

    ul.log(__name__).info("Compute a neighborhood graph of observations.")
    sc.pp.neighbors(adata, n_neighbors=n_neighbors)

    if is_denoise:
        ul.log(__name__).info("Run denoising")
        sc.tl.diffmap(adata)
        sc.pp.neighbors(adata, n_neighbors=n_neighbors, use_rep="X_diffmap")

    if groups not in adata.obs.columns:

        if groups == "louvain":
            ul.log(__name__).info("Run louvain")
            sc.tl.louvain(adata, resolution=resolution)
        else:
            ul.log(__name__).error("The value of the `groups` parameter is 'louvain' or needs to be included in `adata.obs.columns`.")
            raise ValueError("The value of the `groups` parameter is 'louvain' or needs to be included in `adata.obs.columns`.")

    ul.log(__name__).info("Run PAGA")
    sc.tl.paga(adata, groups=groups)

    sc.pl.paga(adata, show=False)
    sc.tl.draw_graph(adata, init_pos="paga")

    if position is not None:
        adata.obsm['X_draw_graph_fr_old'] = adata.obsm['X_draw_graph_fr']
        adata.obsm['X_draw_graph_fr'] = adata.obs[list(position)].values

    if root_cluster is not None:
        adata.uns["iroot"] = np.flatnonzero(adata.obs[groups] == root_cluster)[0]
        sc.tl.dpt(adata)

    if status == 1:
        adata.obsm['lsi'] = adata.obsm.pop('X_pca')
    elif status == 2:
        adata.obsm.pop('X_pca')
    else:
        pass



# -*- coding: UTF-8 -*-

import random
import time
from typing import Union, Tuple, Literal, Optional

from scipy import sparse
from scipy.stats import norm
from tqdm import tqdm
from joblib import Parallel, delayed

import numpy as np
from anndata import AnnData
import pandas as pd
from pandas import DataFrame

from ._matrix_ import (
    matrix_dot_block_storage,
    vector_multiply_block_storage
)

from .. import util as ul
from ..util import (
    matrix_data,
    to_sparse,
    to_dense,
    sparse_matrix,
    dense_data,
    number,
    collection,
    get_index,
    difference_peak_optional
)

__name__: str = "tool_algorithm"

_Affinity = Literal["nearest_neighbors", "rbf", "precomputed", "precomputed_nearest_neighbors", "jaccard"]
_EigenSolver = Literal["arpack", "lobpcg", "amg"]


def sigmoid(data: Union[collection, matrix_data]) -> Union[collection, matrix_data]:
    return 1 / (1 + np.exp(-data))


def tf_idf(data: matrix_data, ri_sparse: bool = True) -> matrix_data:
    """
    TF-IDF transformer
    :param data: Matrix data that needs to be converted;
    :param ri_sparse: (return_is_sparse) Whether to return sparse matrix.
    :return: Matrix processed by TF-IDF.
    """
    from sklearn.feature_extraction.text import TfidfTransformer

    ul.log(__name__).info("TF-IDF transformer")
    transformer = TfidfTransformer()
    tfidf = transformer.fit_transform(to_dense(data, is_array=True))
    return to_sparse(tfidf) if ri_sparse else to_dense(tfidf)


def z_score_normalize(
    data: matrix_data,
    with_mean: bool = True,
    ri_sparse: bool | None = None,
    is_sklearn: bool = False
) -> Union[dense_data, sparse_matrix]:
    """
    Matrix standardization (z-score)
    :param data: Standardized data matrix required.
    :param with_mean: If True, center the data before scaling.
    :param ri_sparse: (return_is_sparse) Whether to return sparse matrix.
    :param is_sklearn: This parameter represents whether to use the sklearn package.
    :return: Standardized matrix.
    """
    ul.log(__name__).info("Matrix z-score standardization")

    if is_sklearn:
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler(with_mean=with_mean)

        if with_mean:
            dense_data_ = to_dense(data, is_array=True)
        else:
            dense_data_ = data

        data = scaler.fit_transform(np.array(dense_data_))
    else:

        if sparse.issparse(data):
            _data_: sparse_matrix = data
            __mean__ = np.mean(_data_.data)
            __std__ = np.std(_data_.data)
            data.data = (_data_.data - __mean__) / (1 if __std__ == 0 else __std__)
            del _data_, __mean__, __std__
        else:
            __mean__ = np.mean(data)
            __std__ = np.std(data)
            data = (data - __mean__) / (1 if __std__ == 0 else __std__)

    return data if ri_sparse is None else (to_sparse(data) if ri_sparse else to_dense(data))


def z_score_marginal(matrix: matrix_data, axis: Literal[0, 1] = 0) -> Tuple[matrix_data, matrix_data]:
    """
    Matrix standardization (z-score, marginal)
    :param matrix: Standardized data matrix required.
    :param axis: Standardize according to which dimension.
    :return: Standardized matrix.
    """
    ul.log(__name__).info("Start marginal z-score")
    matrix = np.matrix(to_dense(matrix))
    # Separate z-score for each element
    __mean__ = np.mean(matrix, axis=axis)
    __std__ = np.std(matrix, axis=axis)
    # Control denominator is not zero
    __std__[__std__ == 0] = 1
    _z_score_ = (matrix - __mean__) / __std__
    ul.log(__name__).info("End marginal z-score")
    return _z_score_, __mean__


def z_score_to_p_value(z_score: matrix_data):
    return 2 * (1 - norm.cdf(abs(z_score)))


def marginal_normalize(matrix: matrix_data, axis: Literal[0, 1] = 0, default: float = 1e-50) -> matrix_data:
    """
    Marginal standardization
    :param matrix: Standardized data matrix required;
    :param axis: Standardize according to which dimension;
    :param default: To prevent division by 0, this value needs to be added to the denominator.
    :return: Standardized data.
    """
    matrix = np.matrix(to_dense(matrix))
    __sum__ = np.sum(matrix, axis=axis)
    return matrix / (__sum__ + default)


def min_max_norm(data: matrix_data, axis: Literal[0, 1, -1] = -1) -> dense_data:
    """
    Calculate min max standardized data
    :param data: input data;
    :param axis: Standardize according to which dimension.
    :return: Standardized data.
    """
    data = to_dense(data, is_array=True)

    # Judgment dimension
    if axis == -1:
        data_extremum = data.max() - data.min()
        if data_extremum == 0:
            data_extremum = 1
        new_data = (data - data.min()) / data_extremum
    elif axis == 0:
        data_extremum = np.array(data.max(axis=axis) - data.min(axis=axis)).flatten()
        data_extremum[data_extremum == 0] = 1
        new_data = (data - data.min(axis=axis).flatten()) / data_extremum
    elif axis == 1:
        data_extremum = np.array(data.max(axis=axis) - data.min(axis=axis)).flatten()
        data_extremum[data_extremum == 0] = 1
        new_data = (data - data.min(axis=axis).flatten()[:, np.newaxis]) / data_extremum[:, np.newaxis]
    else:
        ul.log(__name__).error(
            "The `axis` parameter supports only -1, 0, and 1, while other values will make the `scale` parameter value "
            "equal to 1."
        )
        raise ValueError("The `axis` parameter supports only -1, 0, and 1")

    return new_data


def symmetric_scale(
    data: matrix_data,
    scale: Union[number, collection] = 2.0,
    axis: Literal[0, 1, -1] = -1,
    is_verbose: bool = True
) -> matrix_data:
    """
    Symmetric scale Function
    :param data: input data;
    :param axis: Standardize according to which dimension;
    :param scale: scaling factor.
    :param is_verbose: log information.
    :return: Standardized data
    """

    from scipy import special

    if is_verbose:
        ul.log(__name__).info("Start symmetric scale function")

    # Judgment dimension
    if axis == -1:
        scale = 1 if scale == 0 else scale
        x_data = to_dense(data) / scale
    elif axis == 0:
        scale = to_dense(scale, is_array=True).flatten()
        scale[scale == 0] = 1
        x_data = to_dense(data) / scale
    elif axis == 1:
        scale = to_dense(scale, is_array=True).flatten()
        scale[scale == 0] = 1
        x_data = to_dense(data) / scale[:, np.newaxis]
    else:
        ul.log(__name__).warning(
            "The `axis` parameter supports only -1, 0, and 1, while other values will make the `scale` parameter value "
            "equal to 1."
        )
        x_data = to_dense(data)

    # Record symbol information
    symbol = to_dense(x_data).copy()
    symbol[symbol > 0] = 1
    symbol[symbol < 0] = -1

    # Log1p standardized data
    y_data = np.multiply(x_data, symbol)
    y_data = special.log1p(y_data)
    del x_data

    # Return symbols and make changes and sigmoid mapped data
    z_data = np.multiply(y_data, symbol)

    if is_verbose:
        ul.log(__name__).info("End symmetric scale function")
    return z_data


def mean_symmetric_scale(data: matrix_data, axis: Literal[0, 1, -1] = -1, is_verbose: bool = True) -> matrix_data:
    """
    Calculate the mean symmetric
    :param data: input data;
    :param axis: Standardize according to which dimension.
    :param is_verbose: log information.
    :return: Standardized data after average symmetry.
    """

    # Judgment dimension
    if axis == -1:
        return symmetric_scale(data, np.abs(data).mean(), axis=-1, is_verbose=is_verbose)
    elif axis == 0:
        return symmetric_scale(data, np.abs(data).mean(axis=0), axis=0, is_verbose=is_verbose)
    elif axis == 1:
        return symmetric_scale(data, np.abs(data).mean(axis=1), axis=1, is_verbose=is_verbose)
    else:
        ul.log(__name__).warning("The `axis` parameter supports only -1, 0, and 1")
        raise ValueError("The `axis` parameter supports only -1, 0, and 1")


def coefficient_of_variation(matrix: matrix_data, axis: Literal[0, 1, -1] = 0, default: float = 0) -> Union[
    float, collection]:
    if axis == -1:
        _std_ = np.array(np.std(matrix))
        _mean_ = np.array(np.mean(matrix))

        if _mean_ == 0:
            return default
        else:
            factor = _std_ / _mean_

            if factor == 0:
                return default

            return factor
    else:
        _std_ = np.array(np.std(matrix, axis=axis))
        _mean_ = np.array(np.mean(matrix, axis=axis))
        _mean_[_mean_ == 0] = 1 if default == 0 else default
        # coefficient of variation
        factor = _std_ / _mean_
        factor[_std_ == 0] = default
        return factor


def is_asc_sort(positions_list: list) -> bool:
    """
    Judge whether the site is in ascending order
    :param positions_list: positions list.
    :return: True for ascending order, otherwise False.
    """
    length: int = len(positions_list)

    if length <= 1:
        return True

    tmp = positions_list[0]

    for i in range(1, length):
        if positions_list[i] < tmp:
            return False
        tmp = positions_list[i]

    return True


def lsi(data: matrix_data, n_components: int = 50, is_to_dense: bool = False) -> dense_data:
    """
    SVD LSI
    :param data: input cell feature data;
    :param n_components: Dimensions that need to be reduced to.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Reduced dimensional data (SVD LSI model).
    """

    from sklearn.decomposition import TruncatedSVD

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    if data_x.shape[1] <= n_components:
        ul.log(__name__).info(
            "The features of the data are less than or equal to the `n_components` parameter, ignoring LSI"
        )
        return data_x
    else:
        ul.log(__name__).info("Start LSI")
        svd = TruncatedSVD(n_components=n_components, algorithm='randomized')
        svd_data = svd.fit_transform(data_x)
        ul.log(__name__).info("End LSI")
        return svd_data


def pca(data: matrix_data, n_components: int = 50, is_to_dense: bool = False) -> dense_data:
    """
    PCA
    :param data: input cell feature data;
    :param n_components: Dimensions that need to be reduced to.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Reduced dimensional data.
    """
    from sklearn.decomposition import PCA

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    if data_x.shape[1] <= n_components:
        ul.log(__name__).info(
            "The features of the data are less than or equal to the `n_components` parameter, ignoring PCA"
        )
        return data_x
    else:
        ul.log(__name__).info("Start PCA")
        pca_n = PCA(n_components=n_components)
        pca_data = pca_n.fit_transform(data_x)
        ul.log(__name__).info("End PCA")
        return pca_data


def jaccard_similarity(data: matrix_data, n_jobs: int = -1, is_to_dense: bool = False) -> matrix_data:
    """
    Calculate the Jaccard similarity matrix
    :param data: input cell feature data;
    :param n_jobs: The number of jobs to use for the computation.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    """
    from sklearn.metrics import pairwise_distances
    from sklearn.preprocessing import binarize

    ul.log(__name__).info("Start Jaccard Similarity")
    data = binarize(data, threshold=0.5).astype(bool)

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    jaccard_data = 1 - pairwise_distances(data_x, metric='jaccard', n_jobs=n_jobs)
    ul.log(__name__).info("End Jaccard Similarity")
    return jaccard_data


def spectral_eigenmaps(
    data: matrix_data,
    n_components: int = 30,
    affinity: _Affinity = 'nearest_neighbors',
    eigen_solver: Optional[_EigenSolver] = None,
    n_jobs: int = -1,
    is_to_dense: bool = False
) -> dense_data:
    """
    Spectral Eigenmaps
    :param data: input cell feature data;
    :param n_components: Dimensions that need to be reduced to.
    :param eigen_solver: The eigenvalue decomposition strategy to use.
    :param affinity: method
    :param n_jobs: The number of jobs to use for the computation.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Reduced dimensional data.
    """
    from sklearn.manifold import SpectralEmbedding

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    if data_x.shape[1] <= n_components:
        ul.log(__name__).info(
            "The features of the data are less than or equal to the `n_components` parameter, ignoring Spectral "
            "Eigenmaps."
        )
        return data_x
    else:
        ul.log(__name__).info("Start Spectral Eigenmaps")

        if affinity == 'jaccard':
            affinity_matrix = jaccard_similarity(data=data_x, n_jobs=n_jobs)
            affinity = 'precomputed'
        else:
            affinity_matrix = data_x

        se = SpectralEmbedding(
            n_components=n_components,
            affinity=affinity,
            eigen_solver=eigen_solver,
            n_jobs=n_jobs
        )
        se_data = se.fit_transform(affinity_matrix)
        ul.log(__name__).info("End Spectral Eigenmaps")
        return se_data


def semi_mutual_knn_weight(
    data: matrix_data,
    neighbors: int = 30,
    or_neighbors: int = 1,
    weight: float = 0.1,
    is_for: bool = True,
    is_mknn_fully_connected: bool = True
) -> Tuple[matrix_data, matrix_data]:
    """
    Mutual KNN with weight
    :param data: Input data matrix;
    :param neighbors: The number of nearest neighbors;
    :param or_neighbors: The number of or nearest neighbors;
    :param weight: The weight of interactions or operations;
    :param is_for: Obtain the nearest neighbors of each node from each row of the for loop matrix;
        Setting it to True is very suitable for situations with large samples and insufficient memory.
    :param is_mknn_fully_connected: Is the network of MKNN an all connected graph?
        If the value is True, it ensures that a node is connected to at least the node that is not closest to itself.
        This parameter does not affect the result of SM-KNN (the first result), but only affects the result of
        traditional M-KNN (the second result).
    :return: Adjacency weight matrix
    """
    ul.log(__name__).info("Start semi-mutual KNN")

    if weight < 0 or weight > 1:
        ul.log(__name__).error("The `and_weight` parameter must be between 0 and 1.")
        raise ValueError("The `and_weight` parameter must be between 0 and 1.")

    # Work directly on the sparse matrix to avoid a full dense copy
    if sparse.issparse(data):
        # Keep sparse, set diagonal to 0 efficiently
        data = to_sparse(data).astype(np.float32)
        data.setdiag(0)
        data.eliminate_zeros()
        new_data = data
        del data
    else:
        # Dense case: in-place diagonal zeroing
        new_data = to_dense(data).astype(np.float32)
        del data
        np.fill_diagonal(new_data, 0)

    def _knn_k_(_mat: matrix_data, k: int, info: str = "LOG"):
        n_rows = _mat.shape[0]
        adj = sparse.lil_matrix((n_rows, n_rows), dtype=np.int8)

        ul.log(__name__).info(f"Calculate the k-nearest neighbors of each node. ({info})")

        for i in tqdm(range(n_rows)):
            row = np.array(_mat[i]).ravel()

            if row.size <= k:
                adj[i, :] = 1
                adj[i, i] = 0
                continue

            # Partial sorting
            kth = np.partition(row, -k)[-k]
            mask = row >= kth
            mask[i] = False  # remove self
            adj[i, mask] = 1

        return adj.tocsr()

    def _knn(_mat: matrix_data, k: int, info: str = "LOG") -> matrix_data:
        """
        Return k-nearest-neighbor 0/1 adjacency matrix (int8 to save memory).
        Supports both sparse and dense inputs.
        """

        if sparse.issparse(_mat):
            # Sparse path: sort each row's data to find the k-th largest
            _mat = _mat.tocsr(copy=False)
            return _knn_k_(_mat, k, info)
        else:

            if is_for:
                return _knn_k_(_mat, k, info)
            else:
                # Dense path: vectorized thresholding
                kth_val = np.sort(_mat, axis=1)[:, -(k + 1)]
                adj = (_mat >= kth_val[:, None]).astype(np.int8)
                np.fill_diagonal(adj, 0)
                return adj

    # Compute adjacency matrices for AND/OR logic
    adj_and = _knn(new_data, neighbors, "AND")

    if neighbors == or_neighbors:
        adj_or = adj_and
    else:
        adj_or = _knn(new_data, or_neighbors, "OR")

    # Symmetrize
    if sparse.issparse(adj_and):
        adj_and = adj_and.minimum(adj_and.T)
        adj_or = adj_or.maximum(adj_or.T)
    else:
        adj_and = np.minimum(adj_and, adj_and.T)
        adj_or = np.maximum(adj_or, adj_or.T)

    # Weighted combination, float32 is sufficient
    adj_weight = (1 - weight) * adj_and.astype(np.float32) + weight * adj_or.astype(np.float32)

    # Ensure full connectivity if required
    if is_mknn_fully_connected:
        adj_1nn = _knn(new_data, 1, "ONE")

        if sparse.issparse(adj_1nn):
            adj_1nn = adj_1nn.maximum(adj_1nn.T)
        else:
            adj_1nn = np.maximum(adj_1nn, adj_1nn.T)

        if sparse.issparse(adj_and):
            adj_and = adj_and.maximum(adj_1nn)
        else:
            adj_and = np.maximum(adj_and, adj_1nn)

    ul.log(__name__).info("End semi-mutual KNN")
    return adj_weight, adj_and


def k_means(data: matrix_data, n_clusters: int = 8, is_to_dense: bool = False):
    """
    Perform k-means clustering on data
    :param data: Input data matrix;
    :param n_clusters: The number of clusters to form as well as the number of centroids to generate.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Tags after k-means clustering.
    """
    ul.log(__name__).info("Start K-means cluster")
    from sklearn.cluster import KMeans

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    model = KMeans(n_clusters=n_clusters, n_init="auto")
    model.fit(to_dense(data_x, is_array=True))
    labels = model.labels_
    ul.log(__name__).info("End K-means cluster")
    return labels


def spectral_clustering(
    data: matrix_data,
    n_clusters: int = 8,
    n_components=30,
    eigen_solver="arpack",
    is_to_dense: bool = False
) -> collection:
    """
    Spectral clustering
    :param data: Input data matrix;
    :param n_clusters: The dimension of the projection subspace.
    :param n_components: The dimension of the projection subspace.
    :param eigen_solver: Default use of Nyström approximation.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Tags after spectral clustering.
    """
    ul.log(__name__).info("Start spectral clustering")

    from sklearn.cluster import SpectralClustering

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    model = SpectralClustering(n_clusters=n_clusters, n_components=n_components, eigen_solver=eigen_solver)
    clusters_types = model.fit_predict(data_x)
    ul.log(__name__).info("End spectral clustering")
    return clusters_types


def tsne(data: matrix_data, n_components: int = 2, is_to_dense: bool = False) -> matrix_data:
    """
    T-SNE dimensionality reduction
    :param data: Data matrix that requires dimensionality reduction;
    :param n_components: Dimension of the embedded space.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Reduced dimensional data matrix
    """
    from sklearn.manifold import TSNE

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    _tsne_ = TSNE(n_components=n_components)
    data_tsne = _tsne_.fit_transform(data_x)
    return data_tsne


def umap(
    data: matrix_data,
    n_neighbors: float = 15,
    n_components: int = 2,
    min_dist: float = 0.15,
    is_to_dense: bool = False
) -> matrix_data:
    """
    UMAP dimensionality reduction
    :param data: Data matrix that requires dimensionality reduction;
    :param n_neighbors: float (optional, default 15)
        The size of local neighborhood (in terms of number of neighboring
        sample points) used for manifold approximation. Larger values
        result in more global views of the manifold, while smaller
        values result in more local data being preserved. In general
        values should be in the range 2 to 100;
    :param n_components: The dimension of the space to embed into. This defaults to 2 to
        provide easy visualization, but can reasonably be set to any
        integer value in the range 2 to 100.
    :param min_dist: The effective minimum distance between embedded points. Smaller values
        will result in a more clustered/clumped embedding where nearby points
        on the manifold are drawn closer together, while larger values will
        result on a more even dispersal of points. The value should be set
        relative to the ``spread`` value, which determines the scale at which
        embedded points will be spread out.
    :param is_to_dense: Whether to convert the data into a dense matrix.
    :return: Reduced dimensional data matrix
    """
    import umap as umap_

    data_x = to_dense(data, is_array=True) if is_to_dense else data

    embedding = umap_.UMAP(n_neighbors=n_neighbors, n_components=n_components, min_dist=min_dist).fit_transform(data_x)
    return embedding


def safe_kl_divergence(p: collection, q: collection, epsilon: float = 1e-10):
    """Safe KL divergence calculation to avoid division by zero"""

    # Ensure p and q are probability distributions
    p = to_dense(p, is_array=True).flatten()
    q = to_dense(q, is_array=True).flatten()

    # Smoothing: add small value to q to avoid zeros
    p = p + epsilon
    p = p / np.sum(p)

    q = q + epsilon
    q = q / np.sum(q)

    kl = np.sum(p * np.log(p / q))

    return kl


def kl_divergence(data1: matrix_data, data2: matrix_data) -> float:
    """
    Calculate KL divergence for two data
    :param data1: First data;
    :param data2: Second data.
    :return: KL divergence score
    """
    from scipy import stats

    data1 = to_dense(data1, is_array=True).flatten()
    data2 = to_dense(data2, is_array=True).flatten()

    entropy = stats.entropy(data1, data2)

    if np.isnan(entropy) or np.isinf(entropy):
        entropy = safe_kl_divergence(data1, data2)

    return entropy


def calinski_harabasz(data: matrix_data, labels: collection) -> float:
    """
    The Calinski-Harabasz index is also one of the indicators used to evaluate the quality of clustering models.
    It measures the compactness within the cluster and the separation between clusters in the clustering results. The
    larger the value, the better the clustering effect
    :param data: First data;
    :param labels: Predicted labels for each sample.
    :return:
    """
    from sklearn.metrics import calinski_harabasz_score
    return calinski_harabasz_score(to_dense(data, is_array=True), labels)


def silhouette(data: matrix_data, labels: collection) -> float:
    """
    silhouette
    :param data: An array of pairwise distances between samples, or a feature array;
    :param labels: Predicted labels for each sample.
    :return: index
    """
    from sklearn.metrics import silhouette_score
    return silhouette_score(to_dense(data, is_array=True), labels)


def davies_bouldin(data: matrix_data, labels: collection) -> float:
    """
    Davies-Bouldin index (DBI)
    :param data: A list of ``n_features``-dimensional data points. Each row corresponds to a single data point;
    :param labels: Predicted labels for each sample.
    :return: index
    """
    from sklearn.metrics import davies_bouldin_score
    return davies_bouldin_score(to_dense(data, is_array=True), labels)


def ari(labels_pred: collection, labels_true: collection) -> float:
    """
    ARI (-1, 1)
    :param labels_pred: Predictive labels for clustering;
    :param labels_true: Real labels for clustering.
    :return: index
    """
    from sklearn.metrics import adjusted_rand_score
    return adjusted_rand_score(labels_true, labels_pred)


def ami(labels_pred: collection, labels_true: collection) -> float:
    """
    AMI (0, 1)
    :param labels_pred: Predictive labels for clustering;
    :param labels_true: Real labels for clustering.
    :return: index
    """
    from sklearn.metrics import adjusted_mutual_info_score
    return adjusted_mutual_info_score(labels_true, labels_pred)


def binary_indicator(
    labels_true: collection,
    labels_pred: collection
) -> Tuple[float, float, float, float, float, float, float]:
    """
    Accuracy, Recall, F1, FPR, TPR, AUROC, AUPRC
    :param labels_true: Real labels for clustering;
    :param labels_pred: Predictive labels for clustering.
    :return: Indicators
    """
    from sklearn.metrics import (
        accuracy_score,
        recall_score,
        f1_score,
        roc_curve,
        roc_auc_score,
        average_precision_score
    )

    acc_s = accuracy_score(labels_true, labels_pred)
    rec_s = recall_score(labels_true, labels_pred)
    f1_s = f1_score(labels_true, labels_pred)
    fpr, tpr, thresholds = roc_curve(labels_true, labels_pred)
    auroc_s = roc_auc_score(labels_true, labels_pred)
    auprc_s = average_precision_score(labels_true, labels_pred)
    return acc_s, rec_s, f1_s, fpr, tpr, auroc_s, auprc_s


def euclidean_distances(data1: matrix_data, data2: matrix_data = None, block_size: int = -1) -> matrix_data:
    """
    Calculate the Euclidean distance between two matrices
    :param data1: First data;
    :param data2: Second data (If the second data is empty, it will default to the first data.)
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed
    :return: Data of Euclidean distance.
    """
    ul.log(__name__).info("Start euclidean distances")

    if data2 is None:
        data2 = data1.copy()

    data1 = to_dense(data1)
    data2 = to_dense(data2)
    __data1_sum_sq__ = np.power(data1, 2).sum(axis=1)
    data1_sum_sq = __data1_sum_sq__.reshape((-1, 1))
    data2_sum_sq = __data1_sum_sq__ if data2 is None else np.power(data2, 2).sum(axis=1)
    del __data1_sum_sq__

    distances = data1_sum_sq + data2_sum_sq - 2 * to_dense(
        matrix_dot_block_storage(data1, data2.transpose(), block_size))
    del data1_sum_sq, data2_sum_sq

    distances[distances < 0] = 0.0
    distances = np.sqrt(distances)
    return distances


def _overlap_(regions_sort: DataFrame, variants: DataFrame) -> DataFrame:
    """
    Relate the peak region and variant site
    :param regions_sort: peaks information
    :param variants: variants information
    :return: The variant maps data in the peak region
    """

    columns = ['variant_id', 'index', 'chr', 'position', 'rsId', 'chr_a', 'start', 'end']

    if regions_sort.shape[0] == 0 or variants.shape[0] == 0:
        ul.log(__name__).warning("Data is empty.")
        return pd.DataFrame(columns=columns)

    variants_sort = variants.sort_values(["chr", "position"])[["variant_id", "chr", "position", "rsId"]]

    # Intersect and Sort
    chr_keys: list = list(set(regions_sort["chr"]).intersection(set(variants_sort["chr"])))
    chr_keys.sort()

    variants_chr_type: dict = {}
    variants_position_list: dict = {}

    # Cyclic region chromatin
    for chr_key in chr_keys:
        # variant chr information
        sort_chr_regions_chr = variants_sort[variants_sort["chr"] == chr_key]
        variants_chr_type.update({chr_key: sort_chr_regions_chr})
        variants_position_list.update({chr_key: list(sort_chr_regions_chr["position"])})

    variants_overlap_info_list: list = []

    for index, chr_a, start, end in zip(regions_sort["index"],
                                        regions_sort["chr"],
                                        regions_sort["start"],
                                        regions_sort["end"]):

        # judge chr
        if chr_a in chr_keys:
            # get chr variant
            variants_chr_type_position_list = variants_position_list[chr_a]

            # judge start and end position
            if start <= variants_chr_type_position_list[-1] and end >= variants_chr_type_position_list[0]:
                # get index
                start_index = get_index(start, variants_chr_type_position_list, False)
                end_index = get_index(end, variants_chr_type_position_list, False)

                # Determine whether it is equal, Equality means there is no overlap
                if start_index != end_index:
                    start_index = start_index if isinstance(start_index, int) else start_index[1]
                    end_index = end_index + 1 if isinstance(end_index, int) else end_index[1]

                    if start_index > end_index:
                        ul.log(__name__).error("The end index in the region is greater than the start index.")
                        raise IndexError("The end index in the region is greater than the start index.")

                    variants_chr_type_chr_a = variants_chr_type[chr_a]
                    # get data
                    variants_overlap_info: DataFrame = variants_chr_type_chr_a[start_index:end_index].copy()
                    variants_overlap_info["index"] = index
                    variants_overlap_info["chr_a"] = chr_a
                    variants_overlap_info["start"] = start
                    variants_overlap_info["end"] = end
                    variants_overlap_info.index = (
                        variants_overlap_info["variant_id"].astype(str) + "_"
                        + variants_overlap_info["index"].astype(str)
                    )
                    variants_overlap_info_list.append(variants_overlap_info)

    # merge result
    if len(variants_overlap_info_list) > 0:
        overlap_data: DataFrame = pd.concat(variants_overlap_info_list, axis=0)
    else:
        return pd.DataFrame(columns=columns)

    return overlap_data


def overlap(regions: DataFrame, variants: DataFrame) -> DataFrame:
    """
    Relate the peak region and variant site
    :param regions: peaks information
    :param variants: variants information
    :return: The variant maps data in the peak region
    """
    regions_columns: list = list(regions.columns)

    if "chr" not in regions_columns or "start" not in regions_columns or "end" not in regions_columns:
        ul.log(__name__).error(
            f"The peaks information {regions_columns} in data `adata` must include three columns: `chr`, `start` and "
            f"`end`. (It is recommended to use the `read_sc_atac` method.)"
        )
        raise ValueError(
            f"The peaks information {regions_columns} in data `adata` must include three columns: `chr`, `start` and "
            f"`end`. (It is recommended to use the `read_sc_atac` method.)"
        )

    columns = ['variant_id', 'index', 'chr', 'position', 'rsId', 'chr_a', 'start', 'end']

    if regions.shape[0] == 0 or variants.shape[0] == 0:
        ul.log(__name__).warning("Data is empty.")
        return pd.DataFrame(columns=columns)

    regions = regions.rename_axis("index")
    regions = regions.reset_index()
    # sort
    regions_sort = regions.sort_values(["chr", "start", "end"])[["index", "chr", "start", "end"]]

    return _overlap_(regions_sort, variants)


def overlap_sum(regions: AnnData, variants: dict, trait_info: DataFrame, n_jobs: int = -1) -> AnnData:
    """
    Overlap regional data and mutation data and sum the PP values of all mutations in a region as the values for that
    region.
    :param regions: peaks data
    :param variants: variants data
    :param trait_info: traits information
    :param n_jobs: The maximum number of concurrently running jobs
    :return: overlap data
    """

    start_time = time.perf_counter()

    # Unique feature set
    label_all = regions.var.index.tolist()
    # Peak number
    label_all_size: int = len(label_all)

    # Pre-build a dict of peak indices for O(1) lookup
    label2idx = {lb: i for i, lb in enumerate(label_all)}

    trait_names = trait_info["id"].tolist()
    n_trait = len(trait_names)

    # Check column existence once
    required = {"chr", "start", "end"}

    if not required.issubset(regions.var.columns):
        ul.log(__name__).error(
            f"The peaks information {regions.var.columns} in data `adata` must include three columns: `chr`, `start` "
            f"and `end`. (It is recommended to use the `read_sc_atac` method.)"
        )
        raise ValueError(
            f"The peaks information {regions.var.columns} in data `adata` must include three columns: `chr`, `start` "
            f"and `end`. (It is recommended to use the `read_sc_atac` method.)"
        )

    regions_df = (
        regions.var
        .reset_index()
        .loc[:, ["index", "chr", "start", "end"]]
        .sort_values(["chr", "start", "end"])
    )

    ul.log(__name__).info("Obtain peak-trait/disease matrix. (overlap variant information)")

    # Function to process a single trait
    def _process_trait_(trait_name, col_idx):

        local_data_vals = []
        local_row_indices = []
        local_col_indices = []

        variant: AnnData = variants[trait_name]
        overlap_df: DataFrame = _overlap_(regions_df, variant.obs)

        if overlap_df.empty:
            return local_data_vals, local_row_indices, local_col_indices

        # Sum at once: first group by label and collect variant_id into a list
        label_var_ids = (
            overlap_df
            .groupby("index")["variant_id"]
            .apply(list)
            .reset_index()
        )

        # Traverse each label, sum once for each variant_id list
        for _, row in label_var_ids.iterrows():
            label = row["index"]
            row_idx = label2idx[label]
            var_ids = row["variant_id"]
            # Sum once for all variant_ids in the list, avoiding row-by-row slicing
            matrix_sum = variant[var_ids, :].X.sum(axis=0)

            if np.isscalar(matrix_sum):
                matrix_sum = np.asarray(matrix_sum).reshape(1)

            # Collect non-zero values
            if matrix_sum.size == 1:
                val = float(matrix_sum)
                if val != 0:
                    local_row_indices.append(row_idx)
                    local_col_indices.append(col_idx)
                    local_data_vals.append(val)
            else:
                for t_idx, v in enumerate(matrix_sum):
                    if v != 0:
                        local_row_indices.append(row_idx)
                        local_col_indices.append(col_idx + t_idx)
                        local_data_vals.append(float(v))

        return local_data_vals, local_row_indices, local_col_indices

    # Use Parallel to process traits in parallel
    results = Parallel(n_jobs=n_jobs)(
        delayed(_process_trait_)(trait_name, col_idx) for col_idx, trait_name in tqdm(enumerate(trait_names))
    )

    # Preallocate length to avoid list dynamic expansion
    total = sum(len(ld) for ld, _, _ in results)
    row_indices = np.empty(total, dtype=np.int32)
    col_indices = np.empty(total, dtype=np.int32)
    data_vals = np.empty(total, dtype=np.float32)

    ptr = 0

    for local_data, local_rows, local_cols in results:
        n = len(local_data)
        row_indices[ptr:ptr + n] = local_rows
        col_indices[ptr:ptr + n] = local_cols
        data_vals[ptr:ptr + n] = local_data
        ptr += n

    # Build sparse matrix, then convert to csr format
    overlap_sparse = sparse.csc_matrix(
        (data_vals, (row_indices, col_indices)),
        shape=(label_all_size, n_trait),
        dtype=np.float32
    ).tocsr()

    overlap_adata = AnnData(overlap_sparse, var=trait_info, obs=regions.var)
    overlap_adata.uns["is_overlap"] = True
    overlap_adata.uns["elapsed_time"] = time.perf_counter() - start_time

    return overlap_adata


def calculate_fragment_weighted_accessibility(input_data: dict, block_size: int = -1) -> matrix_data:
    """
    Calculate the initial trait- or disease-related cell score
    :param input_data:
        1. data: Convert the `counts` matrix to the `fragments` matrix using the `scvi.data.reads_to_fragments`
        2. overlap_data: Peaks-traits/diseases data
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed
    :return: Initial TRS
    """

    if "data" not in input_data:
        ul.log(__name__).error("The `data` field needs to be included in parameter `input_data`.")
        raise ValueError("The `data` field needs to be included in parameter `input_data`.")

    if "overlap_data" not in input_data:
        ul.log(__name__).error("The `overlap_data` field needs to be included in parameter `input_data`.")
        raise ValueError("The `overlap_data` field needs to be included in parameter `input_data`.")

    # Processing data
    ul.log(__name__).info("Data pre conversion.")

    matrix = input_data.pop("data")
    overlap_matrix = input_data.pop("overlap_data")

    if sparse.issparse(matrix):
        matrix = matrix.tocsr(copy=False)
    else:
        matrix = to_sparse(matrix)

    if sparse.issparse(overlap_matrix):
        overlap_matrix = overlap_matrix.tocsr(copy=False)
    else:
        overlap_matrix = to_sparse(overlap_matrix)

    matrix.data = matrix.data.astype(np.float32)
    overlap_matrix.data = overlap_matrix.data.astype(np.float32)

    row_sum = np.asarray(matrix.sum(axis=1)).ravel()
    col_sum = np.asarray(matrix.sum(axis=0)).ravel()
    all_sum = row_sum.sum()

    ul.log(__name__).info("Calculate fragment weighted accessibility ===> (numerator)")
    init_score = matrix.dot(overlap_matrix)

    del matrix

    ul.log(__name__).info("Calculate expected counts matrix ===> (numerator)")
    global_scale_data = vector_multiply_block_storage(row_sum, col_sum, block_size=block_size)

    if block_size <= 0:
        global_scale_data = global_scale_data.astype(np.float32)

    del row_sum, col_sum

    global_scale_data /= all_sum * 1.0

    del all_sum

    ul.log(__name__).info("Calculate fragment weighted accessibility ===> (denominator)")
    overlap_matrix = to_dense(overlap_matrix)
    global_scale_data = global_scale_data.dot(overlap_matrix)

    global_scale_data[global_scale_data == 0] = global_scale_data[global_scale_data != 0].min() / 2

    ul.log(__name__).info("Calculate fragment weighted accessibility.")
    init_score = to_dense(init_score)
    init_score /= global_scale_data

    del global_scale_data

    return init_score


def calculate_init_score_weight(
    adata: AnnData,
    da_peaks_adata: AnnData,
    overlap_adata: AnnData,
    layer: Optional[str] = "fragments",
    diff_peak_value: difference_peak_optional = 'emp_effect',
    is_simple: bool = True,
    block_size: int = -1
) -> AnnData:
    """
    Calculate the initial trait- or disease-related cell score with weight.
    :param adata: scATAC-seq data;
    :param da_peaks_adata: Differential peak data;
    :param overlap_adata: Peaks-traits/diseases data;
    :param layer: The layer value of scATAC-seq data;
    :param diff_peak_value: Specify the correction value in peak correction of clustering type differences.
        {'emp_effect', 'bayes_factor', 'emp_prob1', 'all'}
    :param is_simple: True represents not adding unnecessary intermediate variables, only adding the final result. It
        is worth noting that when set to `True`, the `is_ablation` parameter will become invalid, and when set to
        `False`, `is_ablation` will only take effect;
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed
    :return: Initial TRS with weight.
    """

    start_time = time.perf_counter()

    if "is_overlap" not in overlap_adata.uns:
        ul.log(__name__).warning(
            "The `is_overlap` is not in `overlap_data.uns`. "
            "(Suggest using the 'tl.overlap_stum' function to obtain the result.)"
        )

    if "dp_delta" not in da_peaks_adata.uns:
        ul.log(__name__).warning(
            "The `dp_delta` is not in `da_peaks_adata.uns`. "
            "(Suggest using the 'pp.poisson_vi' function to obtain the result.)"
        )

    if layer is not None and layer not in adata.layers:
        ul.log(__name__).error(
            f"The `layer` parameter is empty or one of the element values of `adata.layers` ({adata.layers})."
        )
        raise ValueError(
            f"The `layer` parameter is empty or one of the element values of `adata.layers` ({adata.layers})."
        )

    fragments = adata.layers[layer] if layer is not None else adata.X
    cell_anno = adata.obs
    del adata

    fragments = to_sparse(fragments.astype(np.int32))
    overlap_matrix = to_dense(overlap_adata.X)
    trait_anno = overlap_adata.var
    del overlap_adata

    ul.log(__name__).info("Calculate cell type weight")

    def _get_cluster_weight_(da_matrix: matrix_data):
        _cluster_weight_data_: matrix_data = min_max_norm(da_matrix, axis=0).dot(overlap_matrix)
        return sigmoid(mean_symmetric_scale(_cluster_weight_data_, axis=0, is_verbose=False))

    if diff_peak_value == "emp_effect":
        _cluster_weight_ = _get_cluster_weight_(da_peaks_adata.X)
    elif diff_peak_value == "bayes_factor":
        _cluster_weight_ = _get_cluster_weight_(da_peaks_adata.layers["bayes_factor"])
    elif diff_peak_value == "emp_prob1":
        _cluster_weight_ = _get_cluster_weight_(da_peaks_adata.layers["emp_prob1"])
    elif diff_peak_value == "all":
        _cluster_weight1_ = _get_cluster_weight_(da_peaks_adata.X)
        _cluster_weight2_ = _get_cluster_weight_(da_peaks_adata.layers["bayes_factor"])
        _cluster_weight3_ = _get_cluster_weight_(da_peaks_adata.layers["emp_prob1"])
        _cluster_weight_ = (_cluster_weight1_ + _cluster_weight2_ + _cluster_weight3_) / 3
        del _cluster_weight1_, _cluster_weight2_, _cluster_weight3_
    else:
        ul.log(__name__).error(
            "The `diff_peak_value` parameter only supports one of the {'emp_effect', 'bayes_factor', 'emp_prob1', "
            "'all'} values."
        )
        raise ValueError(
            "The `diff_peak_value` parameter only supports one of the {'emp_effect', 'bayes_factor', 'emp_prob1', "
            "'all'} values."
        )

    overlap_matrix = to_sparse(overlap_matrix.astype(np.float32))

    row_sum = np.asarray(fragments.sum(axis=1)).ravel()
    col_sum = np.asarray(fragments.sum(axis=0)).ravel()
    all_sum = row_sum.sum()

    ul.log(__name__).info("Calculate fragment weighted accessibility ===> (numerator)")
    _init_trs_ncw_ = fragments.dot(overlap_matrix)

    del fragments

    ul.log(__name__).info("Calculate expected counts matrix ===> (numerator)")
    global_scale_data = vector_multiply_block_storage(row_sum, col_sum, block_size=block_size)

    if block_size <= 0:
        global_scale_data = global_scale_data.astype(np.float32)

    del row_sum, col_sum

    global_scale_data /= all_sum * 1.0

    del all_sum

    ul.log(__name__).info("Calculate fragment weighted accessibility ===> (denominator)")
    overlap_matrix = to_dense(overlap_matrix)
    global_scale_data = global_scale_data.dot(overlap_matrix)
    del overlap_matrix

    global_scale_data[global_scale_data == 0] = global_scale_data[global_scale_data != 0].min() / 2

    ul.log(__name__).info("Calculate fragment weighted accessibility.")
    _init_trs_ncw_ = to_dense(_init_trs_ncw_)
    _init_trs_ncw_ /= global_scale_data

    del global_scale_data

    da_peaks_adata.obsm["cluster_weight"] = to_dense(_cluster_weight_, is_array=True)
    del _cluster_weight_

    ul.log(__name__).info("Broadcasting the weight factor to the cellular level")
    _cell_type_weight_ = np.zeros((cell_anno.shape[0], da_peaks_adata.obsm["cluster_weight"].shape[1]),
                                  dtype=np.float32)

    cluster_series = cell_anno["clusters"]

    for cluster in da_peaks_adata.obs_names:
        mask = cluster_series == cluster
        _cell_type_weight_[mask, :] = da_peaks_adata[cluster, :].obsm["cluster_weight"].flatten().astype(np.float32)

    ul.log(__name__).info("Calculate initial trait relevance scores")
    _init_trs_weight_ = np.multiply(_init_trs_ncw_, _cell_type_weight_)

    if hasattr(_init_trs_weight_, "A"):
        _init_trs_weight_ = _init_trs_weight_.A

    init_trs_adata = AnnData(_init_trs_weight_, obs=cell_anno, var=trait_anno)
    del _init_trs_weight_

    if not is_simple:

        if hasattr(_init_trs_ncw_, "A"):
            _init_trs_ncw_ = _init_trs_ncw_.A

        init_trs_adata.layers["init_trs_ncw"] = _init_trs_ncw_
        init_trs_adata.layers["cell_type_weight"] = to_sparse(_cell_type_weight_)
        init_trs_adata.uns["cluster_weight_factor"] = da_peaks_adata.obsm["cluster_weight"]

    del _init_trs_ncw_, _cell_type_weight_

    init_trs_adata.uns["is_sample"] = is_simple
    init_trs_adata.uns["elapsed_time"] = time.perf_counter() - start_time
    return init_trs_adata


def adaptive_gamma_knn(data: matrix_data, k: int = 10):
    """
    Adaptive gamma parameter based on k-nearest neighbors
    :param data: Data matrix (n_samples, n_features)
    :param k: Number of neighbors, usually select 5-20
    :return: Gamma value for each sample (n_samples,)
    """

    from sklearn.neighbors import NearestNeighbors

    # Calculate the distance from each point to its k-th nearest neighbor
    knn = NearestNeighbors(n_neighbors=k + 1).fit(data)  # +1 because it includes itself
    distances, _ = knn.kneighbors(data)

    # Take the distance of the k-th nearest neighbor (index k, because 0 is itself)
    kth_distances = distances[:, k]

    # Avoid division by zero (if the distance is 0, set it to a very small value)
    kth_distances[kth_distances == 0] = np.finfo(float).eps

    # Calculate local gamma: gamma = 1 / (2 * sigma^2), where sigma = kth_distance
    gammas = 1.0 / (kth_distances ** 2)

    return gammas


def obtain_cell_cell_network(
    adata: AnnData,
    k: int = 30,
    or_k: int = 1,
    weight: float = 0.1,
    kernel: Literal["laplacian", "gaussian"] = "gaussian",
    local_k: int = 10,
    gamma: Optional[Union[float, collection]] = None,
    is_simple: bool = True
) -> AnnData:
    """
    Calculate cell-cell correlation
    :param adata: scATAC-seq data;
    :param k: When building an M-KNN network, the number of nodes connected by each node (and);
    :param or_k: When building an M-KNN network, the number of nodes connected by each node (or);
    :param weight: The weight of interactions or operations;
    :param local_k: Determining the number of neighbors for the adaptive kernel;
    :param kernel: Determine the kernel function to be used;
    :param gamma: If None, it defaults to the adaptive value obtained through the local information of
        parameter `local_k`. Otherwise, it should be strictly positive;
    :param is_simple: True represents not adding unnecessary intermediate variables, only adding the final result.
        It is worth noting that when set to `True`, the `is_ablation` parameter will become invalid, and when set to
        `False`, `is_ablation` will only take effect;
    :return: Cell similarity data.
    """

    start_time = time.perf_counter()

    from sklearn.metrics.pairwise import laplacian_kernel, rbf_kernel

    # data
    if "poisson_vi" not in adata.uns.keys():
        ul.log(__name__).error(
            "`poisson_vi` is not in the `adata.uns` dictionary, and the scATAC-seq data needs to be processed through "
            "the `poisson_vi` function."
        )
        raise ValueError(
            "`poisson_vi` is not in the `adata.uns` dictionary, and the scATAC-seq data needs to be processed through "
            "the `poisson_vi` function."
        )

    if kernel not in ["laplacian", "gaussian"]:
        ul.log(__name__).error("Parameter `kernel` only supports two values, `laplacian` and `gaussian`.")
        raise ValueError("Parameter `kernel` only supports two values, `laplacian` and `gaussian`.")

    if local_k <= 0:
        ul.log(__name__).error("The `local_k` parameter must be a natural number greater than 0.")
        raise ValueError("The `local_k` parameter must be a natural number greater than 0.")

    _latent_name_ = "latent" if adata.uns["poisson_vi"]["latent_name"] is None \
        else adata.uns["poisson_vi"]["latent_name"]

    latent = adata.obsm[_latent_name_].astype(np.float32)
    del _latent_name_
    cell_anno = adata.obs
    del adata

    if gamma is None:
        gamma = adaptive_gamma_knn(latent, k=local_k)

    if kernel == "kernel":
        ul.log(__name__).info("Laplacian kernel")
        cell_affinity = laplacian_kernel(latent, gamma=gamma).astype(np.float32)
    else:
        ul.log(__name__).info("Gaussian (RBF) kernel")
        cell_affinity = rbf_kernel(latent, gamma=gamma).astype(np.float32)

    # Define KNN network
    cell_mutual_knn_weight, cell_mutual_knn = semi_mutual_knn_weight(
        cell_affinity,
        neighbors=k,
        or_neighbors=or_k,
        weight=weight
    )

    if is_simple:
        del cell_mutual_knn

    # cell-cell graph
    cc_data: AnnData = AnnData(cell_mutual_knn_weight, var=cell_anno, obs=cell_anno)
    cc_data.layers["cell_affinity"] = cell_affinity

    if not is_simple:
        cc_data.layers["cell_mutual_knn"] = to_sparse(cell_mutual_knn)

    cc_data.uns["elapsed_time"] = time.perf_counter() - start_time

    return cc_data


def perturb_data(data: collection, percentage: float) -> collection:
    """
    Randomly perturbs the positions of a percentage of data.
    :param data: List of data elements to be perturbed.
    :param percentage: Percentage of data to be perturbed.
    :return: Perturbed data list.
    """

    if percentage <= 0 or percentage > 1:
        raise ValueError("The value of the `percentage` parameter must be greater than 0 and less than or equal to 1.")

    new_data = data.copy()
    num_elements = len(new_data)
    num_to_perturb = int(num_elements * percentage)

    # Select random indices to perturb
    indices_to_perturb = random.sample(range(num_elements), num_to_perturb)

    # Swap elements at selected indices with other random elements
    for index in indices_to_perturb:
        swap_index = random.choice([i for i in range(num_elements) if i != index])
        new_data[index], new_data[swap_index] = new_data[swap_index], new_data[index]

    return new_data


def add_bernoulli_fluctuation_noise(
    counts_matrix: matrix_data,
    noise_level: float = 0.1
) -> matrix_data:
    """
    Add Bernoulli fluctuation noise to the counts matrix (add 1 with probability noise_level)

    Parameters
    ----------
    counts_matrix : matrix_data
        Input counts matrix
    noise_level : float, default 0.1
        Noise level, i.e., the probability of randomly adding 1 (range: 0.0 - 1.0)

    Returns
    -------
    matrix_data
        Matrix after adding noise
    """
    if noise_level < 0 or noise_level > 1:
        ul.log(__name__).error("The value of the `noise_level` parameter must be greater than 0 and less than 1.")
        raise ValueError("The value of the `noise_level` parameter must be greater than 0 and less than 1.")

    if noise_level == 0:
        return counts_matrix.copy()

    noise = np.random.binomial(
        n=1,
        p=noise_level,
        size=counts_matrix.shape
    ).astype(counts_matrix.dtype)

    return counts_matrix + noise



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Literal, Optional, Union, Any

import pandas as pd
from anndata import AnnData
from pandas import DataFrame
from tqdm import tqdm

from ._scatter_ import scatter_trait, scatter_atac
from ._heat_map_ import heatmap_annotation
from ._pie_ import pie_trait
from ._box_ import box_trait
from ._barcode_ import barcode_trait
from ._violin_ import violin_trait
from ._bar_ import bar_trait
from ._kde_ import kde
from ._graph_ import communities_graph
from ._radar_ import radar_trait

from .. import util as ul
from ..preprocessing import adata_group, adata_map_df
from ..util import path, collection

__name__: str = "plot_core"


def group_heatmap(
    adata: AnnData,
    layer: str = None,
    dir_name: str = "heatmap",
    row_name: str = "labels",
    col_name: str = "clusters",
    clusters: str = "clusters",
    cluster_method: str = "average",
    cluster_metric: str = "correlation",
    x_label_rotation: float = 245,
    group_method: collection = ("mean", "sum", "median"),
    row_cluster: bool = True,
    col_cluster: bool = True,
    row_index: list = None,
    col_index: list = None,
    is_sort: bool = True,
    width: float = 4,
    height: float = 2,
    axis: Literal[0, 1] = 1,
    plot_output: str = None,
    show: bool = True
) -> None:
    """
    Generate a heatmap for the specified AnnData data grouping.
    :param adata: input data;
    :param layer: Specify the matrix to be processed;
    :param dir_name: Folder name for generating heatmaps;
    :param row_name: The title of the row coordinates in the figure;
    :param col_name: The title of the column coordinates in the figure;
    :param clusters: The column that need to be grouped.
    :param cluster_method: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param cluster_metric: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param x_label_rotation: The degree of rotation for row names. Effective when `row_names_side` is `true`;
    :param group_method: The method of grouping strategy supports the following 5 types and their combinations.
        The five methods are {"mean", "sum", "median", "max", "min"}.
    :param row_cluster: Set to `true`, perform row clustering;
    :param col_cluster: Set to `true`, perform column clustering;
    :param row_index: Sort the specified row order to display the graph;
    :param col_index: Sort the specified column order to display the graph;
    :param is_sort: If set to true, when displaying the heatmap, the row and column names are sorted and displayed;
        When `row_index` or `col_index` has a value, this parameter becomes invalid.
    :param width: The width of the file or image;
    :param height: The height of the file or image;
    :param axis: Which dimension is used for grouping. {1: adata.obs, 0: adata.var};
    :param layer: Layer for processing data;
    :param plot_output: Output file for image saving;
    :param show: If true, display the image;
    :return:
    """

    # The relationship between cluster and trait
    adata_cluster: AnnData = adata_group(
        adata,
        clusters,
        extra_column=None if col_name == clusters else col_name,
        method=group_method,
        layer=layer,
        axis=axis
    )

    if row_index is not None:
        adata_cluster = adata_cluster[row_index, :].copy()
        is_sort = False

    if col_index is not None:
        adata_cluster = adata_cluster[:, col_index].copy()
        is_sort = False

    new_path: path = os.path.join(plot_output, dir_name) if plot_output is not None else None
    # create path
    if plot_output is not None:
        ul.file_method(__name__).makedirs(new_path)

    # plot
    def _show_plot_(_label_: str, _layer_: str = None) -> None:
        try:
            heatmap_annotation(
                adata_cluster.T,
                layer=_layer_,
                row_name=row_name,
                col_name=col_name,
                width=width,
                height=height,
                label="Mean TRS",
                row_legend=True,
                x_label_rotation=x_label_rotation,
                cluster_method=cluster_method,
                cluster_metric=cluster_metric,
                row_cluster=row_cluster,
                col_cluster=col_cluster,
                is_sort=is_sort,
                row_show_names=True,
                col_show_names=True,
                x_name="Cell type",
                y_name="Trait",
                show=show,
                output=os.path.join(new_path, f"{layer}_{_layer_}.pdf") if new_path is not None else None
            )
        except Exception as e:
            ul.log(__name__).warning(f"Changing clustering parameters {e.args}")
            heatmap_annotation(
                adata_cluster.T,
                row_name=row_name,
                col_name=col_name,
                width=width,
                height=height,
                label="Mean TRS",
                cluster_method="median",
                cluster_metric="euclidean",
                row_legend=True,
                x_label_rotation=x_label_rotation,
                row_cluster=row_cluster,
                col_cluster=col_cluster,
                is_sort=is_sort,
                row_show_names=True,
                col_show_names=True,
                x_name="Cell type",
                y_name="Trait",
                show=show,
                output=os.path.join(new_path, f"{layer}_{_layer_}.pdf") if new_path is not None else None
            )

    layers = adata_cluster.layers

    _show_plot_("Mean TRS", "mean")

    if "sum" in layers:
        _show_plot_("Sum TRS", "sum")

    if "median" in layers:
        _show_plot_("Median TRS", "median")

    if "max" in layers:
        _show_plot_("Max TRS", "max")

    if "min" in layers:
        _show_plot_("Min TRS", "min")


def map_df_plot(
    adata: AnnData,
    layer: str = None,
    trait_cluster_map: dict = None,
    clusters: str = "clusters",
    feature_name: str = "feature",
    width: float = 8,
    y_name: str = "value",
    column: str = "value",
    show: bool = True,
    plot_output: path = None
) -> None:
    # create path
    new_path: path = os.path.join(
        plot_output,
        f"{layer}" if layer is not None else feature_name
    ) if plot_output is not None else None

    # create path
    if plot_output is not None:
        ul.file_method(__name__).makedirs(new_path)

    # create data
    adata_df: DataFrame = adata_map_df(adata, column=column, layer=layer)

    # pie plot
    if trait_cluster_map is not None:
        pie_trait(
            adata_df,
            trait_cluster_map=trait_cluster_map,
            clusters=clusters,
            title=feature_name,
            show=show,
            output=new_path if plot_output is not None else None
        )

    # box plot
    box_trait(
        adata_df,
        y_name=y_name,
        value=column,
        clusters=clusters,
        width=width,
        title=feature_name,
        show=show,
        output=new_path if plot_output is not None else None
    )

    # rank plot
    barcode_trait(
        adata_df,
        sort_column=column,
        clusters=clusters,
        title=feature_name,
        show=show,
        output=new_path if plot_output is not None else None
    )

    # violin plot
    violin_trait(
        adata_df,
        y_name=y_name,
        value=column,
        width=width,
        clusters=clusters,
        title=feature_name,
        show=show,
        output=new_path if plot_output is not None else None
    )


def complete_ratio(
    adata: AnnData,
    layer: str = None,
    column: str = "value",
    extra_columns: collection = None,
    clusters: str = "clusters"
) -> DataFrame:
    # create data
    adata_df: DataFrame = adata_map_df(adata, column=column, layer=layer)

    clusters_group = adata_df.groupby(["id", clusters], as_index=False).size()
    value_group = adata_df.groupby(["id", clusters, column], as_index=False).size()
    new_value_group = value_group.merge(clusters_group, on=["id", clusters], how="left")

    if extra_columns is not None:
        extra_columns = list(extra_columns)
        extra_columns.extend(["id", clusters])
        new_value_group = new_value_group.merge(adata_df[extra_columns].drop_duplicates(), on=["id", clusters], how="left")

    # Completion
    id_list = list(set(new_value_group["id"]))
    clusters_list = list(set(new_value_group[clusters]))
    value_list = [1.0, 0.0]
    total_size = len(id_list) * len(clusters_list) * len(value_list)

    if total_size != new_value_group.shape[0]:
        new_value_group_index = (
            new_value_group["id"].astype(str) + "_"
            + new_value_group[clusters].astype(str) + "_"
            + new_value_group[column].astype(int).astype(str)
        )
        new_value_group.index = new_value_group_index
        new_value_group_index = list(new_value_group_index)

        trait_df: DataFrame = pd.DataFrame(columns=new_value_group.columns)

        # [id clusters  `column`  size_x size_y `extra_columns`]
        for _id_ in tqdm(id_list):
            for _clusters_ in clusters_list:
                for _value_ in value_list:

                    # At this point, it means that the enrichment effect is 1, while the non enrichment effect is 0,
                    # so it does not exist during grouping and needs to be added here
                    if (_id_ + "_" + _clusters_ + "_" + str(int(_value_))) not in new_value_group_index:
                        exit_value = 0 if int(_value_) == 1 else 1
                        exit_index = _id_ + "_" + _clusters_ + "_" + str(exit_value)
                        exit_data = new_value_group[new_value_group.index == exit_index]
                        exit_data.loc[exit_index, column] = _value_
                        exit_data.loc[exit_index, "size_x"] = 0
                        exit_data.index = [_id_ + "_" + _clusters_ + "_" + str(int(_value_))]
                        trait_df = pd.concat((trait_df, exit_data), axis=0)

        new_value_group = pd.concat((trait_df, new_value_group), axis=0)

    new_value_group["rate"] = new_value_group["size_x"] / new_value_group["size_y"]

    return new_value_group


def rate_bar_plot(
    adata: AnnData,
    layer: str = None,
    trait_name: str = "All",
    dir_name: str = "feature",
    column: str = "value",
    clusters: str = "clusters",
    color: Tuple = ("#2e6fb7", "#f7f7f7"),
    legend: Tuple = ("Enrichment", "Conservative"),
    x_name: str = "Cell type",
    y_name: str = "Enrichment ratio",
    clusters_sort: Optional[list] = None,
    text_color: str = "#000205",
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    rotation: float = 65,
    title: str = None,
    text_left_move: float = 0.15,
    y_limit: Tuple[float, float] = (0, 1),
    plot_output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:

    if dir_name is not None:
        # create path
        new_path = os.path.join(plot_output, f"{dir_name}_{layer}") if plot_output is not None else None
        # create path
        if plot_output is not None:
            ul.file_method(__name__).makedirs(new_path)
    else:
        plot_output = None
        new_path = None

    # create data
    new_value_group = complete_ratio(adata=adata, layer=layer, column=column, clusters=clusters)

    bar_trait(
        trait_df=new_value_group,
        value="rate",
        clusters=clusters,
        title=title,
        x_name=x_name,
        y_name=y_name,
        clusters_sort=clusters_sort,
        trait_name=trait_name,
        color=color,
        legend=legend,
        text_color=text_color,
        width=width,
        height=height,
        bottom=bottom,
        rotation=rotation,
        text_left_move=text_left_move,
        y_limit=y_limit,
        output=new_path if plot_output is not None else None,
        show=show,
        close=close,
        **kwargs
    )


def rate_circular_bar_plot(
    adata: AnnData,
    layer: str = None,
    trait_name: str = "All",
    dir_name: str = "feature",
    column: str = "value",
    clusters: str = "clusters",
    color: Union[collection, str] = None,
    clusters_sort: Optional[list] = None,
    width: float = 2,
    height: float = 2,
    rotation: float = 25,
    title: str = None,
    value_top: float = 0.1,
    text_top: float = 1.2,
    is_fixed: bool = False,
    is_angle: bool = True,
    y_limit: Tuple = (-0.5, 1),
    y_axis_scale: Tuple = (0, 1),
    plot_output: path = None,
    show: bool = True,
    close: bool = False
) -> None:

    if dir_name is not None:
        # create path
        new_path = os.path.join(plot_output, f"{dir_name}_{layer}") if plot_output is not None else None
        # create path
        if plot_output is not None:
            ul.file_method(__name__).makedirs(new_path)
    else:
        plot_output = None
        new_path = None

    # create data
    if color is not None and isinstance(color, str):
        new_value_group = complete_ratio(adata=adata, layer=layer, column=column, extra_columns=[color], clusters=clusters)
    else:
        new_value_group = complete_ratio(adata=adata, layer=layer, column=column, clusters=clusters)

    new_value_group = new_value_group[new_value_group[column] == 1].copy()

    radar_trait(
        trait_df=new_value_group,
        value="rate",
        clusters=clusters,
        title=title,
        clusters_sort=clusters_sort,
        trait_name=trait_name,
        color=color,
        width=width,
        height=height,
        rotation=rotation,
        value_top=value_top,
        text_top=text_top,
        is_fixed=is_fixed,
        is_angle=is_angle,
        y_limit=y_limit,
        y_axis_scale=y_axis_scale,
        output=new_path if plot_output is not None else None,
        show=show,
        close=close
    )


def init_score_plot(
    init_score: AnnData,
    plot_output: str,
    clusters: str = "clusters",
    plot_columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    trait_cluster_map: dict = None,
    width: float = 4
) -> None:
    # plot
    ul.file_method(__name__).makedirs(os.path.join(plot_output, "trait_cell"))
    layers: list = list(init_score.layers)

    scatter_trait(init_score, output=os.path.join(plot_output, "trait_cell"), title="", columns=plot_columns)
    scatter_trait(init_score, output=os.path.join(plot_output, "trait_cell"), title="", layers=layers, columns=plot_columns)

    try:
        kde(
            init_score,
            axis=0,
            title="Initialization score with weights",
            output=os.path.join(plot_output, f"init_score_weight_kde")
        )
    except Exception as e:
        ul.log(__name__).warning(f"{e}")
        ul.log(__name__).warning(f"Can try to improve the version of `seaborn`. `pip install --upgrade seaborn`")

    group_heatmap(adata=init_score, dir_name="heatmap", col_name=clusters, clusters=clusters, plot_output=plot_output)
    map_df_plot(
        adata=init_score,
        trait_cluster_map=trait_cluster_map,
        feature_name="cell",
        clusters=clusters,
        y_name="Initial score",
        column="value",
        width=width,
        plot_output=plot_output
    )

    # Display of Heat Map and Violin Map
    for layer in layers:
        kde(init_score, axis=0, layer=layer, title=layer, output=os.path.join(plot_output, f"{layer}_kde"))
        group_heatmap(
            adata=init_score,
            layer=layer,
            dir_name="heatmap",
            col_name=clusters,
            clusters=clusters,
            plot_output=plot_output
        )
        map_df_plot(
            adata=init_score,
            layer=layer,
            trait_cluster_map=trait_cluster_map,
            feature_name="cell",
            clusters=clusters,
            y_name="Initial score",
            column="value",
            width=width,
            plot_output=plot_output
        )


def cell_cell_plot(
    cc_data: AnnData,
    plot_output: str,
    clusters: str = "clusters",
    is_graph: bool = False
) -> None:
    layers: list = list(cc_data.layers)
    # heatmap
    heatmap_annotation(
        cc_data,
        layer="cell_affinity",
        col_name=clusters,
        row_name=clusters,
        row_legend=True,
        row_anno_label=True,
        col_anno_label=True,
        output=os.path.join(plot_output, "cc_heatmap.png")
    )
    heatmap_annotation(
        cc_data,
        col_name=clusters,
        row_name=clusters,
        row_legend=True,
        row_anno_label=True,
        col_anno_label=True,
        output=os.path.join(plot_output, "cc_heatmap_mknn_weight.png")
    )

    if "cell_mutual_knn" in layers:
        heatmap_annotation(
            cc_data,
            layer="cell_mutual_knn",
            col_name=clusters,
            row_name=clusters,
            row_legend=True,
            row_anno_label=True,
            col_anno_label=True,
            output=os.path.join(plot_output, "cc_heatmap_mknn.png")
        )

    if is_graph:
        # Cell-cell network
        communities_graph(
            adata=cc_data,
            clusters=clusters,
            labels=cc_data.obs[clusters],
            output=os.path.join(plot_output, "cc_weight_graph")
        )

        if "cell_mutual_knn" in layers:
            communities_graph(
                adata=cc_data,
                layer="cell_mutual_knn",
                clusters=clusters,
                labels=cc_data.obs[clusters],
                output=os.path.join(plot_output, "cc_graph")
            )


def data_plot(
    init_score: AnnData,
    cc_data: AnnData,
    trs: AnnData,
    plot_output: str,
    clusters: str = "clusters",
    plot_columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    trait_cluster_map: dict = None,
    is_graph: bool = False,
    width: float = 4
) -> None:
    # scatter
    scatter_atac(trs, clusters=clusters, output=os.path.join(plot_output, "scATAC_cluster.pdf"), columns=plot_columns)

    kde(
        init_score, axis=0, title="Initial score with weight (scale)",
        output=os.path.join(plot_output, "init_score_weight_scale_kde")
    )

    trs_plot(
        trs=trs,
        plot_output=plot_output,
        clusters=clusters,
        plot_columns=plot_columns,
        trait_cluster_map=trait_cluster_map,
        width=width
    )

    cell_cell_plot(
        cc_data=cc_data,
        plot_output=plot_output,
        clusters=clusters,
        is_graph=is_graph
    )


def trs_plot(
    trs: AnnData,
    plot_output: str,
    clusters: str = "clusters",
    plot_columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    trait_cluster_map: dict = None,
    heatmap_width: float = 4,
    heatmap_height: float = 2,
    width: float = 4
) -> None:
    trs_layers: list = list(trs.layers)

    # plot
    ul.file_method(__name__).makedirs(os.path.join(plot_output, "kde"))
    ul.file_method(__name__).makedirs(os.path.join(plot_output, "trait_cell"))
    scatter_trait(trs, output=os.path.join(plot_output, "trait_cell"), title="", columns=plot_columns)
    scatter_trait(
        trs,
        output=os.path.join(plot_output, "trait_cell"),
        title="",
        layers=trs_layers,
        columns=plot_columns
    )

    group_heatmap(
        adata=trs,
        dir_name="heatmap",
        col_name=clusters,
        clusters=clusters,
        width=heatmap_width,
        height=heatmap_height,
        plot_output=plot_output
    )
    map_df_plot(
        adata=trs,
        feature_name="init_score_weight",
        trait_cluster_map=trait_cluster_map,
        clusters=clusters,
        y_name="Cell score",
        column="value",
        width=width,
        plot_output=plot_output
    )

    try:
        kde(
            adata=trs,
            axis=0,
            title="Initialization score with weights",
            output=os.path.join(plot_output, "kde", f"init_score_weight_kde")
        )
    except Exception as e:
        ul.log(__name__).warning(f"{e}")
        ul.log(__name__).warning(f"Can try to improve the version of `seaborn`. `pip install --upgrade seaborn`")

    # Display of Heat Map and Violin Map
    for layer in trs_layers:
        group_heatmap(
            adata=trs,
            layer=layer,
            dir_name="heatmap",
            col_name=clusters,
            clusters=clusters,
            width=heatmap_width,
            height=heatmap_height,
            plot_output=plot_output
        )
        map_df_plot(
            adata=trs,
            layer=layer,
            feature_name="",
            trait_cluster_map=trait_cluster_map,
            clusters=clusters,
            y_name="Cell score",
            column="value",
            width=width,
            plot_output=plot_output
        )

        try:
            kde(
                adata=trs,
                axis=0,
                layer=layer,
                title=f"Distribution of {layer}",
                output=os.path.join(plot_output, "kde", f"{layer}_kde")
            )
        except Exception as e:
            ul.log(__name__).warning(f"{e}")
            ul.log(__name__).warning(f"Can try to improve the version of `seaborn`. `pip install --upgrade seaborn`")

        if layer.startswith("run_en"):
            rate_bar_plot(
                adata=trs,
                dir_name="enrichment",
                clusters=clusters,
                layer=layer,
                width=width,
                column="value",
                plot_output=plot_output
            )
